"""This module contains all the methods to obtain information about the
model uncertainty.

"""


def rfunc_uncertainty(ml, rfunc, alpha):
    """

    Parameters
    ----------
    ml: pastas.Model
        Pastas model instance
    rfunc: pastas.rfunc.RfuncBase
        Pastas response function
    alpha: float
        float between 0 and 1 that indicates the

    Returns
    -------

    """

    return series



# coding=utf-8
"""This module contains all the response functions available in Pastas.

More information on how to write a response class can be found `here
 <http://pastas.readthedocs.io/en/latest/developers.html>_`.

Routines in Module
------------------
Fully supported and tested routines in this module are:

- .. class:: Gamma
- .. class:: Exponential
- .. class:: Hantush
- .. class:: One

TODO
----
- Test Polder response function

"""

import numpy as np
from pandas import DataFrame
from scipy.special import gammainc, gammaincinv, k0, exp1, erfc, lambertw

__all__ = ["Gamma", "Exponential", "Hantush", "One"]


class RfuncBase:
    _name = "RfuncBase"

    def __init__(self, up, meanstress, cutoff):
        self.up = up
        # Completely arbitrary number to prevent divsion by zero
        if meanstress < 1e-8 and meanstress > 0:
            meanstress = 1e-8
        elif meanstress < 0 and up is True:
            meanstress = meanstress * -1
        self.meanstress = meanstress
        self.cutoff = cutoff
        self.tmax = 0

    def set_parameters(self, name):
        pass

    def get_tmax(self, p, cutoff=0.99):
        """Method to get the response time for a certain cutoff

        Parameters
        ----------
        p:  numpy.array
            numpy array with the parameters.
        cutoff: float, optional
            float between 0 and 1. Default is 0.99.

        Returns
        -------
        tmax: float
            Number of days when 99% of the response has passen, when the
            cutoff is chosen at 0.99.

        """
        pass

    def step(self, p, dt=1, cutoff=0.99):
        """Method to return the step funtion.

        Parameters
        ----------
        p: numpy.array
            numpy array with the parameters.
        dt: float
            timestep as a multiple of of day.
        cutoff: float, optional
            float between 0 and 1. Default is 0.99.

        Returns
        -------
        s: numpy.array
            Array with the step response.
        """
        pass

    def block(self, p, dt=1, cutoff=0.99):
        """Method to return the block funtion.

        Parameters
        ----------
        p: numpy.array
            numpy array with the parameters.
        dt: float
            timestep as a multiple of of day.
        cutoff: float, optional
            float between 0 and 1. Default is 0.99.

        Returns
        -------
        s: numpy.array
            Array with the block response.
        """
        s = self.step(p, dt, cutoff)
        return np.append(s[0], s[1:] - s[:-1])


class Gamma(RfuncBase):
    """Gamma response function with 3 parameters A, a, and n.

    Parameters
    ----------
    up: bool, optional
        indicates whether a positive stress will cause the head to go up
        (True, default) or down (False)
    meanstress: float
        mean value of the stress, used to set the initial value such that
        the final step times the mean stress equals 1
    cutoff: float
        percentage after which the step function is cut off. default=0.99.

    Notes
    -----

    .. math::
        step(t) = A * Gammainc(n, t / a)

    """
    _name = "Gamma"

    def __init__(self, up=True, meanstress=1, cutoff=0.99):
        RfuncBase.__init__(self, up, meanstress, cutoff)
        self.nparam = 3

    def set_parameters(self, name):
        parameters = DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        if self.up:
            parameters.loc[name + '_A'] = (1 / self.meanstress, 0,
                                           100 / self.meanstress, 1, name)
        else:
            parameters.loc[name + '_A'] = (-1 / self.meanstress,
                                           -100 / self.meanstress, 0, 1, name)
        # if n is too small, the length of the response function is close to zero
        parameters.loc[name + '_n'] = (1, 0.1, 10, 1, name)
        parameters.loc[name + '_a'] = (10, 0.01, 5000, 1, name)
        return parameters

    def get_tmax(self, p, cutoff=None):
        if cutoff is None:
            cutoff = self.cutoff
        return gammaincinv(p[1], cutoff) * p[2]

    def gain(self, p):
        return p[0]

    def step(self, p, dt=1, cutoff=0.99):
        if isinstance(dt, np.ndarray):
            t = dt
        else:
            self.tmax = max(self.get_tmax(p, cutoff), 3 * dt)
            t = np.arange(dt, self.tmax, dt)

        s = p[0] * gammainc(p[1], t / p[2])
        return s


class Exponential(RfuncBase):
    """Exponential response function with 2 parameters: A and a.

        Parameters
        ----------
        up: bool, optional
            indicates whether a positive stress will cause the head to go up
            (True, default) or down (False)
        meanstress: float
            mean value of the stress, used to set the initial value such that
            the final step times the mean stress equals 1
        cutoff: float
            percentage after which the step function is cut off. default=0.99.

        Notes
        -----
        .. math::
            step(t) = A * (1 - exp(-t / a))

        """
    _name = "Exponential"

    def __init__(self, up=True, meanstress=1, cutoff=0.99):
        RfuncBase.__init__(self, up, meanstress, cutoff)
        self.nparam = 2

    def set_parameters(self, name):
        parameters = DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        if self.up:
            parameters.loc[name + '_A'] = (
                1 / self.meanstress, 0, 100 / self.meanstress, 1, name)
        else:
            parameters.loc[name + '_A'] = (
                -1 / self.meanstress, -100 / self.meanstress, 0, 1, name)
        parameters.loc[name + '_a'] = (10, 0.01, 5000, 1, name)
        return parameters

    def get_tmax(self, p, cutoff=None):
        if cutoff is None:
            cutoff = self.cutoff
        return -p[1] * np.log(1 - cutoff)

    def gain(self, p):
        return p[0]

    def step(self, p, dt=1, cutoff=0.99):
        if isinstance(dt, np.ndarray):
            t = dt
        else:
            self.tmax = max(self.get_tmax(p, cutoff), 3 * dt)
            t = np.arange(dt, self.tmax, dt)
        s = p[0] * (1.0 - np.exp(-t / p[1]))
        return s


class Hantush(RfuncBase):
    """ The Hantush well function.

    Parameters
    ----------
    up: bool, optional
        indicates whether a positive stress will cause the head to go up
        (True, default) or down (False)
    meanstress: float
        mean value of the stress, used to set the initial value such that
        the final step times the mean stress equals 1
    cutoff: float
        percentage after which the step function is cut off. default=0.99.

    Notes
    -----
    Parameters are ..math::
        rho = r / lambda and cS

    References
    ----------
    .. [1] Hantush, M. S., & Jacob, C. E. (1955). Non‐steady radial flow in an
        infinite leaky aquifer. Eos, Transactions American Geophysical Union,
        36(1), 95-100.

    .. [2] Veling, E. J. M., & Maas, C. (2010). Hantush well function
    revisited. Journal of hydrology, 393(3), 381-388.

    .. [3] Von Asmuth, J. R., Maas, K., Bakker, M., & Petersen, J. (2008).
    Modeling time series of ground water head fluctuations subjected to
    multiple stresses. Ground Water, 46(1), 30-40.

    """
    _name = "Hantush"

    def __init__(self, up=False, meanstress=1, cutoff=0.99):
        RfuncBase.__init__(self, up, meanstress, cutoff)
        self.nparam = 3

    def set_parameters(self, name):
        parameters = DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        if self.up:
            parameters.loc[name + '_A'] = (
                1 / self.meanstress, 0, 100 / self.meanstress, 1, name)
        else:
            parameters.loc[name + '_A'] = (
                -1 / self.meanstress, -100 / self.meanstress, 0, 1, name)
        parameters.loc[name + '_rho'] = (1, 0.0001, 10, 1, name)
        parameters.loc[name + '_cS'] = (100, 1e-3, 1e3, 1, name)
        return parameters

    def get_tmax(self, p, cutoff=None):
        # approximate formula for tmax
        if cutoff is None:
            cutoff = self.cutoff
        rho = p[1]
        cS = p[2]
        k0rho = k0(rho)
        return lambertw(1 / ((1 - cutoff) * k0rho)).real * cS

    def gain(self, p):
        return p[0]

    def step(self, p, dt=1, cutoff=0.99):
        rho = p[1]
        cS = p[2]
        k0rho = k0(rho)
        if isinstance(dt, np.ndarray):
            t = dt
        else:
            self.tmax = max(self.get_tmax(p, cutoff), 3 * dt)
            t = np.arange(dt, self.tmax, dt)
        tau = t / cS
        tau1 = tau[tau < rho / 2]
        tau2 = tau[tau >= rho / 2]
        w = (exp1(rho) - k0rho) / (exp1(rho) - exp1(rho / 2))
        F = np.zeros_like(tau)
        F[tau < rho / 2] = w * exp1(rho ** 2 / (4 * tau1)) - (w - 1) * exp1(
            tau1 + rho ** 2 / (4 * tau1))
        F[tau >= rho / 2] = 2 * k0rho - w * exp1(tau2) + (w - 1) * exp1(
            tau2 + rho ** 2 / (4 * tau2))
        return p[0] * F / (2 * k0rho)


class Polder(RfuncBase):
    """The function of Polder, for a river in a confined aquifer,
    overlain by an aquitard with aquiferous ditches.

    References
    ----------
    .. [2] http://grondwaterformules.nl/index.php/formules/waterloop/deklaag
    -met-sloten

    """
    _name = "Polder"

    def __init__(self, up=True, meanstress=1, cutoff=0.99):
        RfuncBase.__init__(self, up, meanstress, cutoff)
        self.nparam = 3

    def set_parameters(self, name):
        parameters = DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        a_init = 1
        b_init = 0.1
        c_init = 1 / np.exp(-2 * a_init) / self.meanstress
        parameters.loc[name + '_a'] = (a_init, 0, 100, 1, name)
        parameters.loc[name + '_b'] = (b_init, 0, 10, 1, name)
        if self.up:
            parameters.loc[name + '_c'] = (c_init, 0, c_init * 100, 1, name)
        else:
            parameters.loc[name + '_c'] = (-c_init, -c_init * 100, 0, 1, name)
        return parameters

    def get_tmax(self, p, cutoff=None):
        # TODO: find tmax from cutoff, below is just an opproximation
        return 4 * p[0] / p[1] ** 2

    def gain(self, p):
        # TODO: check line below
        return p[2]

    def step(self, p, dt=1, cutoff=0.99):
        if isinstance(dt, np.ndarray):
            t = dt
        else:
            self.tmax = max(self.get_tmax(p, cutoff), 3 * dt)
            t = np.arange(dt, self.tmax, dt)
        s = p[2] * self.polder_function(p[0], p[1] * np.sqrt(t))
        return s

    def polder_function(self, x, y):
        s = .5 * np.exp(2 * x) * erfc(x / y + y) + \
            .5 * np.exp(-2 * x) * erfc(x / y - y)
        return s


class One(RfuncBase):
    """Dummy class for Constant. Returns 1

    """
    _name = "One"

    def __init__(self, up, meanstress, cutoff):
        RfuncBase.__init__(self, up, meanstress, cutoff)
        self.nparam = 1

    def set_parameters(self, name):
        parameters = DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        if self.up:
            parameters.loc[name + '_d'] = (1, 0, 100, 1, name)
        else:
            parameters.loc[name + '_d'] = (-1, -100, 0, 1, name)
        return parameters

    def gain(self, p):
        return p[0]

    def step(self, p, dt=1):
        if isinstance(dt, np.ndarray):
            return p[0] * np.ones(len(dt))
        else:
            return p[0] * np.ones(1)

    def block(self, p, dt=1):
        return p[0] * np.ones(1)



import numpy as np
from pandas import Series, Timedelta, DataFrame
import matplotlib.pyplot as plt


def acf(x, lags=None, bin_method='gaussian', bin_width=None, max_gap=np.inf,
        min_obs=10, output="acf"):
    """Method to calculate the autocorrelation function for irregular
    timesteps based on the slotting technique. Different methods (kernels)
    to bin the data are available.

    Parameters
    ----------
    x: pandas.Series
        Pandas Series containig the values to calculate the
        cross-correlation for. The index has to be a Pandas.DatetimeIndex
    lags: numpy.array, optional
        numpy array containing the lags in days for which the
        cross-correlation if calculated. [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
        12, 13, 14, 30, 61, 90, 120, 150, 180, 210, 240, 270, 300, 330, 365]
    bin_method: str, optional
        method to determine the type of bin. Options are "gaussian" (default),
        sinc and rectangle.
    bin_width: float, optional
        number of days used as the width for the bin to calculate the
        correlation. By default these values are chosed based on the
        bin_method.
    max_gap: float, optional
        Maximum timestep gap in the data. All timesteps above this gap value
        are not used for calculating the average timestep. This can be
        helpfull when there is a large gap in the data that influences the
        average timestep.

    Returns
    -------
    CCF: pandas.Series
        The Cross-correlation function.

    References
    ----------
    Rehfeld, K., Marwan, N., Heitzig, J., Kurths, J. (2011). Comparison
    of correlation analysis techniques for irregularly sampled time series.
    Nonlinear Processes in Geophysics. 18. 389-404. 10.5194 pg-18-389-2011.

    Examples
    --------
    acf = ps.stats.ccf(x, y, bin_method="gaussian")

    """
    C = ccf(x=x, y=x, lags=lags, bin_method=bin_method, bin_width=bin_width,
            max_gap=max_gap, min_obs=min_obs, output=output)
    C.name = "ACF"  # TODO Rename
    return C


def ccf(x, y, lags=None, bin_method='gaussian', bin_width=None,
        max_gap=np.inf, min_obs=10, output="ccf"):
    """Method to calculate the cross-correlation function for irregular
    timesteps based on the slotting technique. Different methods (kernels)
    to bin the data are available.

    Parameters
    ----------
    x, y: pandas.Series
        Pandas Series containig the values to calculate the
        cross-correlation for. The index has to be a Pandas.DatetimeIndex
    lags: numpy.array, optional
        numpy array containing the lags in days for which the
        cross-correlation if calculated. [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
        12, 13, 14, 30, 61, 90, 120, 150, 180, 210, 240, 270, 300, 330, 365]
    bin_method: str, optional
        method to determine the type of bin. Options are "gaussian" (default),
        sinc and rectangle.
    bin_width: float, optional
        number of days used as the width for the bin to calculate the
        correlation. By default these values are chosed based on the
        bin_method.
    max_gap: float, optional
        Maximum timestep gap in the data. All timesteps above this gap value
        are not used for calculating the average timestep. This can be
        helpfull when there is a large gap in the data that influences the
        average timestep.

    Returns
    -------
    CCF: pandas.Series
        The Cross-correlation function.

    References
    ----------
    Rehfeld, K., Marwan, N., Heitzig, J., Kurths, J. (2011). Comparison
    of correlation analysis techniques for irregularly sampled time series.
    Nonlinear Processes in Geophysics. 18. 389-404. 10.5194 pg-18-389-2011.

    Examples
    --------
    acf = ps.stats.ccf(x, y, bin_method="gaussian")

    """
    # prepare the time indices for x and y
    dt_x = x.index.to_series().diff().values / Timedelta(1, "D")
    dt_x[0] = 0.0
    dt_x_mu = dt_x[dt_x < max_gap].mean()  # Deal with big gaps if present
    t_x = np.cumsum(dt_x)

    dt_y = y.index.to_series().diff().values / Timedelta(1, "D")
    dt_y[0] = 0.0
    dt_y_mu = dt_y[dt_y < max_gap].mean()
    t_y = np.cumsum(dt_y)

    dt_mu = max(dt_x_mu, dt_y_mu)

    # Create matrix with time differences
    t1, t2 = np.meshgrid(t_x, t_y)
    t = np.abs(np.subtract(t1, t2))  # absolute values

    # Normalize the values and create numpy arrays
    x = (x.values - x.values.mean()) / x.values.std()
    y = (y.values - y.values.mean()) / y.values.std()

    # Create matrix for covariances
    xy = np.outer(y, x)

    if lags is None:  # Default lags in Days, log-scale between 0 and 365.
        lags = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14, 30, 61, 90, 120,
                150, 180, 210, 240, 270, 300, 330, 365]

    # Remove lags that cannot be determined because lag < dt_min
    u, i = np.unique(dt_x, return_counts=True)
    dt_x_min = u[Series(i, u).cumsum() >= min_obs][0]
    u, i = np.unique(dt_y, return_counts=True)
    dt_y_min = u[Series(i, u).cumsum() >= min_obs][0]

    dt_min = min(dt_x_min, dt_y_min)
    # dt_min = min(dt_x[1:].min(), dt_y[1:].min())

    lags = np.array([float(lag) for lag in lags if lag >= dt_min or lag == 0])

    # Delete to free memory
    del (x, y, dt_x, dt_y, t1, t2, t_x, t_y)

    # Select appropriate bin_width, default depend on bin_method
    if bin_width is None:
        options = {"rectangle": 0.5, "sinc": 1, "gaussian": 0.25}
        bin_width = np.ones_like(lags) * options[bin_method] * dt_mu
    elif type(bin_width) is float:
        bin_width = np.ones_like(lags)
    else:
        bin_width = [0.5, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 5, 5, 5,
                     2, 2, 2, 2, 2, 2, 2, 2]

    # Select the binning method to calculate the cross-correlation
    if bin_method == "rectangle":
        a = np.zeros_like(t, dtype=float)
        kernel_func = lambda d, h: np.less_equal(np.abs(d, out=a), h,
                                                 out=a).astype(int)
    elif bin_method == "gaussian":
        a = np.zeros_like(t, dtype=float)

        def kernel_func(d, h):
            den1 = -2 * h ** 2  # denominator 1
            den2 = np.sqrt(2 * np.pi * h)  # denominator 2
            return np.exp(np.square(d, out=a) / den1, out=a) / den2
    elif bin_method == "sinc":
        kernel_func = lambda d, h: np.sin(np.pi * h * d) / (np.pi * h * d)
    else:
        raise NotImplementedError("bin_method %s is not implemented." %
                                  bin_method)

    # Pre-allocate an array to speed up all numpy methods
    UDCF = np.zeros_like(lags, dtype=float)
    M = np.zeros_like(lags, dtype=float)
    d = np.zeros_like(t, dtype=float)

    for i, k in enumerate(lags):
        # Construct the kernel for the lag
        np.subtract(t, k, out=d)
        h = bin_width[i]
        b = kernel_func(d, h)
        c = np.multiply(xy, b, out=d)  # Element-wise multiplication
        UDCF[i] = np.sum(c)
        M[i] = np.sum(b)

    DCF = UDCF / M

    C = Series(data=DCF, index=lags, name="CCF")
    CCF = C / C.abs().max()

    if output == "full":
        CCFstd = np.sqrt((np.cumsum(UDCF) - M * DCF) ** 2) / (M - 1)
        CCF = DataFrame(data={"CCF": CCF.values, "stderr": CCFstd}, index=lags)

    CCF.index.name = "Lags (Days)"
    return CCF



"""
This file contains the plotting functionalities that are available for Pastas.

Examples
--------
    ml.plot.decomposition()

"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MultipleLocator
from scipy.stats import probplot

from .decorators import model_tmin_tmax
from .stats import acf

import logging

logger = logging.getLogger(__name__)


class Plotting:
    def __init__(self, ml):
        self.ml = ml  # Store a reference to the model class

    def __repr__(self):
        msg = "This module contains all the built-in plotting options that " \
              "are available."
        return msg

    @model_tmin_tmax
    def plot(self, tmin=None, tmax=None, oseries=True, simulation=True,
             **kwargs):
        """Make a plot of the observed and simulated series.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        oseries: bool, optional
            True to plot the observed time series.
        simulation: bool, optional
            True to plot the simulated time series.

        Returns
        -------
        ax: matplotlib.axes
            matplotlib axes with the simulated and optionally the observed
            timeseries.

        """
        ax = plt.subplot(**kwargs)
        ax.set_title("Results of {}".format(self.ml.name))

        if oseries:
            o = self.ml.observations(tmin=tmin, tmax=tmax)
            o_nu = self.ml.oseries.series.drop(o.index)
            if not o_nu.empty:
                # plot parts of the oseries that are not used in grey
                o_nu.plot(linestyle='', marker='.', color='0.5', label='',
                          ax=ax)
            o.plot(linestyle='', marker='.', color='k', ax=ax)

        if simulation:
            sim = self.ml.simulate(tmin=tmin, tmax=tmax)
            sim.plot(ax=ax)
        plt.xlim(tmin, tmax)
        plt.ylabel("Groundwater levels [meter]")
        plt.legend()
        plt.tight_layout()
        return ax

    @model_tmin_tmax
    def results(self, tmin=None, tmax=None, figsize=(10, 8), **kwargs):
        """Plot different results in one window to get a quick overview.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        figsize: tuple, optional
            tuple of size 2 to determine the figure size in inches.

        Returns
        -------
        matplotlib.axes

        """
        # Number of rows to make the figure with
        rows = 3 + len(self.ml.stressmodels)
        fig = plt.figure(figsize=figsize, **kwargs)
        # Main frame
        ax1 = plt.subplot2grid((rows, 3), (0, 0), colspan=2, rowspan=2)
        o = self.ml.observations(tmin=tmin, tmax=tmax)
        o_nu = self.ml.oseries.series.drop(o.index)
        if not o_nu.empty:
            # plot parts of the oseries that are not used in grey
            o_nu.plot(ax=ax1, linestyle='', marker='.', color='0.5', label='',
                      x_compat=True)
        o.plot(ax=ax1, linestyle='', marker='.', color='k', x_compat=True)
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        sim.plot(ax=ax1, x_compat=True)
        ax1.legend(loc=(0, 1), ncol=3, frameon=False)
        ax1.set_ylim(min(o.min(), sim.loc[tmin:tmax].min()),
                     max(o.max(), sim.loc[tmin:tmax].max()))
        ax1.minorticks_off()

        # Residuals and noise
        ax2 = plt.subplot2grid((rows, 3), (2, 0), colspan=2, sharex=ax1)
        ax2.axhline(0.0, color='k', linestyle='--')
        res = self.ml.residuals(tmin=tmin, tmax=tmax)
        res.plot(ax=ax2, sharex=ax1, color='k', x_compat=True)
        if self.ml.settings["noise"] and self.ml.noisemodel:
            noise = self.ml.noise(tmin=tmin, tmax=tmax)
            noise.plot(ax=ax2, sharex=ax1, x_compat=True)
        ax2.legend(loc=(0, 1), ncol=3, frameon=False)
        ax2.minorticks_off()

        # Stats frame
        ax3 = plt.subplot2grid((rows, 3), (0, 2), rowspan=3)
        ax3.set_title('Model Information', loc='left')

        # Add a row for each stressmodel
        for i, sm in enumerate(self.ml.stressmodels.keys(), start=3):
            ax = plt.subplot2grid((rows, 3), (i, 0), colspan=2, sharex=ax1)
            contrib = self.ml.get_contribution(sm, tmin=tmin, tmax=tmax)
            contrib.plot(ax=ax, sharex=ax1, x_compat=True)
            title = [stress.name for stress in self.ml.stressmodels[sm].stress]
            plt.title("Stresses:%s" % title, loc="right")
            ax.legend(loc=(0, 1), ncol=3, frameon=False)
            if i == 3:
                sharex = None
            else:
                sharex = axb
            axb = plt.subplot2grid((rows, 3), (i, 2), sharex=sharex)
            self.ml.get_step_response(sm).plot(ax=axb)
            ax.minorticks_off()

        ax1.set_xlim(tmin, tmax)

        plt.tight_layout(pad=0.0)

        # Draw parameters table
        parameters = self.ml.parameters.copy()
        parameters['name'] = parameters.index
        cols = ["name", "optimal", "stderr"]
        parameters = parameters.loc[:, cols]
        for name, vals in parameters.loc[:, cols].iterrows():
            parameters.loc[name, "optimal"] = '{:.2f}'.format(vals.optimal)
            stderr_perc = np.abs(np.divide(vals.stderr, vals.optimal) * 100)
            parameters.loc[name, "stderr"] = '{:.1f}{}'.format(stderr_perc,
                                                               "\u0025")
        ax3.axis('off')
        # loc='upper center'
        ax3.table(bbox=(0., 0., 1.0, 1.0), cellText=parameters.values,
                  colWidths=[0.5, 0.25, 0.25], colLabels=cols)

        return fig.axes

    @model_tmin_tmax
    def decomposition(self, tmin=None, tmax=None, ytick_base=True, split=True,
                      figsize=(10, 8), axes=None, name=None,
                      return_warmup=False, **kwargs):
        """Plot the decomposition of a time-series in the different stresses.

        Parameters
        ----------
        split
        ytick_base: Boolean or float
            Make the ytick-base constant if True, set this base to float if float
        **kwargs:
            Optional arguments for the subplots method

        Returns
        -------
        axes: list of matplotlib.axes

        """
        o = self.ml.observations(tmin=tmin, tmax=tmax)

        # determine the simulation
        sim = self.ml.simulate(tmin=tmin, tmax=tmax,
                               return_warmup=return_warmup)
        if name is not None:
            sim.name = name
        series = [sim]
        names = ['']

        # determine the influence of the different stresses
        for name in self.ml.stressmodels.keys():
            nstress = len(self.ml.stressmodels[name].stress)
            if split and nstress > 1:
                for istress in range(nstress):
                    contrib = self.ml.get_contribution(name, tmin=tmin,
                                                       tmax=tmax,
                                                       istress=istress,
                                                       return_warmup=return_warmup)
                    series.append(contrib)
                    names.append(contrib.name)
            else:
                contrib = self.ml.get_contribution(name, tmin=tmin,
                                                   tmax=tmax,
                                                   return_warmup=return_warmup)
                series.append(contrib)
                names.append(contrib.name)

        if self.ml.transform:
            series.append(
                self.ml.get_transform_contribution(tmin=tmin, tmax=tmax))
            names.append(self.ml.transform.name)

        # determine ylim for every graph, to scale the height
        ylims = [
            (min([sim[tmin:tmax].min(), o[tmin:tmax].min()]),
             max([sim[tmin:tmax].max(), o[tmin:tmax].max()]))]
        for contrib in series[1:]:
            hs = contrib[tmin:tmax]
            if hs.empty:
                if contrib.empty:
                    ylims.append((0.0, 0.0))
                else:
                    ylims.append((contrib.min(), hs.max()))
            else:
                ylims.append((hs.min(), hs.max()))
        height_ratios = [
            0.0 if np.isnan(ylim[1] - ylim[0]) else ylim[1] - ylim[0] for ylim
            in ylims]

        if axes is None:
            # open a new figure
            fig, axes = plt.subplots(len(series), sharex=True, figsize=figsize,
                                     gridspec_kw={
                                         'height_ratios': height_ratios},
                                     **kwargs)
            axes = np.atleast_1d(axes)
            o_label = o.name
            set_axes_properties = True
        else:
            assert len(axes) == len(series), 'Makes sure the number of axes ' \
                                             'equals the number of series'
            fig = axes[0].figure
            o_label = ''
            set_axes_properties = False

        # plot simulation and observations in top graph
        o_nu = self.ml.oseries.series.drop(o.index)
        if not o_nu.empty:
            # plot parts of the oseries that are not used in grey
            o_nu.plot(linestyle='', marker='.', color='0.5', label='',
                      markersize=2, ax=axes[0], x_compat=True)
        o.plot(linestyle='', marker='.', color='k', label=o_label,
               markersize=3, ax=axes[0], x_compat=True)
        sim.plot(ax=axes[0], x_compat=True)
        if set_axes_properties:
            axes[0].set_title('Observations vs simulation')
            axes[0].set_ylim(ylims[0])
        axes[0].grid(which='both')
        axes[0].legend(ncol=3, frameon=False)

        if ytick_base and set_axes_properties:
            if isinstance(ytick_base, bool):
                # determine the ytick-spacing of the top graph
                yticks = axes[0].yaxis.get_ticklocs()
                if len(yticks) > 1:
                    ytick_base = yticks[1] - yticks[0]
                else:
                    ytick_base = None
            axes[0].yaxis.set_major_locator(
                MultipleLocator(base=ytick_base))

        # plot the influence of the stresses
        for i, contrib in enumerate(series[1:], start=1):
            contrib.plot(ax=axes[i], x_compat=True)
            if set_axes_properties:
                if ytick_base:
                    # set the ytick-spacing equal to the top graph
                    axes[i].yaxis.set_major_locator(
                        MultipleLocator(base=ytick_base))

                axes[i].set_title(names[i])
                axes[i].set_ylim(ylims[i])
            axes[i].grid(which='both')
            axes[i].minorticks_off()
        if set_axes_properties:
            axes[0].set_xlim(tmin, tmax)
        fig.tight_layout(pad=0.0)

        return axes

    @model_tmin_tmax
    def diagnostics(self, tmin=None, tmax=None):
        """Plot a window that helps in diagnosing basic model assumptions.

        Parameters
        ----------
        tmin
        tmax

        Returns
        -------
        matplotlib.axes

        """
        if self.ml.settings["noise"]:
            res = self.ml.noise(tmin=tmin, tmax=tmax)
        else:
            res = self.ml.residuals(tmin=tmin, tmax=tmax)

        shape = (2, 3)
        ax = plt.subplot2grid(shape, (0, 0), colspan=2, rowspan=1)
        ax.set_title(res.name)
        res.plot(ax=ax)

        ax1 = plt.subplot2grid(shape, (1, 0), colspan=2, rowspan=1)
        ax1.set_ylabel('Autocorrelation')
        conf = 1.96 / np.sqrt(res.index.size)
        r = acf(res)

        ax1.axhline(conf, linestyle='--', color="dimgray")
        ax1.axhline(-conf, linestyle='--', color="dimgray")
        ax1.stem(r.index, r.values, basefmt="gray")
        ax1.set_xlim(r.index.min(), r.index.max())
        ax1.set_xlabel("Lag (Days)")

        ax2 = plt.subplot2grid(shape, (0, 2), colspan=1, rowspan=1)
        res.hist(bins=20, ax=ax2)

        ax3 = plt.subplot2grid(shape, (1, 2), colspan=1, rowspan=1)
        probplot(res, plot=ax3, dist="norm", rvalue=True)

        c = ax.get_lines()[0]._color
        ax3.get_lines()[0].set_color(c)

        plt.tight_layout(pad=0.0)
        return plt.gca()

    def block_response(self, stressmodels=None, **kwargs):
        """Plot the block response for a specific stressmodels.

        Parameters
        ----------
        stressmodels: list, optional
            List with the stressmodels to plot the block response for.

        Returns
        -------
        matplotlib.axes
            matplotlib axes instance.

        """
        if not stressmodels:
            stressmodels = self.ml.stressmodels.keys()

        legend = []

        ax = plt.subplot(**kwargs)

        for name in stressmodels:
            if hasattr(self.ml.stressmodels[name], 'rfunc'):
                self.ml.get_block_response(name).plot(ax=ax)
                legend.append(name)
            else:
                logger.warning("Stressmodel {} not in stressmodels "
                               "list.".format(name))

        plt.xlim(0)
        plt.xlabel("Time [days]")
        plt.legend(legend)
        return ax

    def step_response(self, stressmodels=None, **kwargs):
        """Plot the block response for a specific stressmodels.

        Parameters
        ----------
        stressmodels: list, optional
            List with the stressmodels to plot the block response for.

        Returns
        -------
        matplotlib.axes
            matplotlib axes instance.

        """
        if not stressmodels:
            stressmodels = self.ml.stressmodels.keys()

        legend = []

        ax = plt.subplot(**kwargs)

        for name in stressmodels:
            if hasattr(self.ml.stressmodels[name], 'rfunc'):
                self.ml.get_step_response(name).plot(ax=ax)
                legend.append(name)
            else:
                logger.warning("Stressmodel {} not in stressmodels "
                               "list.".format(name))

        plt.xlim(0)
        plt.xlabel("Time [days]")
        plt.legend(legend)
        return ax

    @model_tmin_tmax
    def stresses(self, tmin=None, tmax=None, cols=1, split=True, sharex=True,
                 figsize=(10, 8), **kwargs):
        """This method creates a graph with all the stresses used in the
         model.

        Parameters
        ----------
        tmin
        tmax
        cols: int
            number of columns used for plotting.

        Returns
        -------
        axes: matplotlib.axes
            matplotlib axes instance.

        """
        stresses = []

        for name in self.ml.stressmodels.keys():
            nstress = len(self.ml.stressmodels[name].stress)
            if split and nstress > 1:
                for istress in range(nstress):
                    stress = self.ml.get_stress(name, istress=istress)
                    stresses.append(stress)
            else:
                stress = self.ml.get_stress(name)
                if isinstance(stress, list):
                    stresses.extend(stress)
                else:
                    stresses.append(stress)

        rows = len(stresses)
        rows = -(-rows // cols)  # round up with out additional import

        fig, axes = plt.subplots(rows, cols, sharex=sharex, figsize=figsize,
                                 **kwargs)

        if hasattr(axes, "flatten"):
            axes = axes.flatten()
        else:
            axes = [axes]

        for ax, stress in zip(axes, stresses):
            stress.plot(ax=ax)
            ax.legend([stress.name], loc=2)

        plt.xlim(tmin, tmax)
        fig.tight_layout(pad=0.0)

        return axes

    @model_tmin_tmax
    def contributions_pie(self, tmin=None, tmax=None, ax=None, **kwargs):
        """Make a pie chart of the contributions. This plot is based on the
        TNO Groundwatertoolbox.

        Parameters
        ----------
        tmin
        tmax
        ax: matplotlib.axes, optional
            Axes to plot the pie chart on. A new figure and axes will be
            created of not providided.
        kwargs: dict, optional
            The keyword arguments are passed on to plt.pie.

        Returns
        -------
        ax: matplotlib.axes

        """
        if ax is None:
            _, ax = plt.subplots()

        frac = []
        for name in self.ml.stressmodels.keys():
            frac.append(np.abs(self.ml.get_contribution(name, tmin=tmin,
                                                        tmax=tmax)).sum())

        evp = self.ml.stats.evp(tmin=tmin) / 100
        frac = np.array(frac) / sum(frac) * evp
        frac = frac.tolist()
        frac.append(1 - evp)
        frac = np.array(frac)
        labels = list(self.ml.stressmodels.keys())
        labels.append("Unexplained")
        ax.pie(frac, labels=labels, autopct='%1.1f%%', startangle=90,
               wedgeprops=dict(width=1, edgecolor='w'), **kwargs)
        ax.axis('equal')
        return ax



import pastas.read as read
import pastas.stats as stats
from .model import Model
from .noisemodels import NoiseModel, NoiseModel2
from .project import Project
from .read import read_meny, read_dino, read_knmi, read_waterbase
from .rfunc import Gamma, Exponential, Hantush, Polder, One
from .solver import LmfitSolve, LeastSquares, DESolve, MarkSolver
from .stressmodels import StressModel, StressModel2, Constant, FactorModel
from .timeseries import TimeSeries
from .transform import ThresholdTransform
from .version import __version__



"""This module contains the mapping methods for Pastas Projects.

Raoul Collenteur, 2018 - Artesia Water

"""

import matplotlib.pyplot as plt


class Map:
    def __init__(self, mls):
        """

        Parameters
        ----------
        mls: pastas.Project
            Pastas project

        """
        self.mls = mls

    def parameter(self, parameter, models=None, param_value="optimal", s=30,
                  show_nan=True, label=False, **kwargs):
        """Plot the value of a parameter.

        Parameters
        ----------
        parameter: str
            String with the name of the parameter to plot.
        models: list, optional
            List of the models top plot the parameter for. By default,
            all models are plotted.
        param_value: str, optional
             String with the parameter kind to be plotted. Any column name
             of the Model parameter DataFrame (e.g. optimal, stderr)
        s: int, optional
            Size of the marker.
        show_nan: bool, optional
            Show nan-values (default is True), which occur when the
            parameter is not in the model.
        label: bool, optional
            Label the points by the model name, default is False.
        kwargs: dict, optional
            Any arguments that are passed on the the "values" method".

        Returns
        -------
        sc: matplotlib.axes
            The axes are returned.

        """
        values = self.mls.get_parameters(parameters=[parameter], models=models,
                                         param_value=param_value)

        sc = self.values(values, models, show_nan, label, s, **kwargs)

        return sc

    def statistic(self, statistic, models=None, s=30, show_nan=True,
                  label=False, **kwargs):
        """Plot the value of a parameter.

        Parameters
        ----------
        statistic: str
            String with the name of the statistic to plot. Must be exactly
            similar to the methods name in the pastas.stats module.
        models: list, optional
            List of the models top plot the parameter for. By default,
            all models are plotted.
        s: int, optional
            Size of the marker.
        show_nan: bool, optional
            Show nan-values (default is True), which occur when the
            parameter is not in the model.
        label: bool, optional
            Label the points by the model name, default is False.
        kwargs: dict, optional
            Any arguments that are passed on the the "values" method".

        Returns
        -------
        sc: matplotlib.axes
            The axes are returned.

        """
        values = self.mls.get_statistics(statistics=[statistic], models=models)

        sc = self.values(values, models, show_nan, label, s, **kwargs)

        return sc

    def values(self, values, models=None, show_nan=True, label=False, s=30,
               **kwargs):
        """Plot the value of a parameter.

        Parameters
        ----------
        values: pandas.Series
            Series with the values to plot, the index should be the model
            names as are used in mls.models.keys()
        models: list, optional
            List of the models top plot the parameter for. By default,
            all models are plotted.
        param_value: str, optional
             String with the parameter kind to be plotted. Any column name
             of the Model parameter DataFrame (e.g. optimal, stderr)
        s: int, optional
            Size of the marker.
        show_nan: bool, optional
            Show nan-values (default is True), which occur when the
            parameter is not in the model.
        label: bool, optional
            Label the points by the model name, default is False.
        kwargs: dict, optional
            Any arguments that are passed on the the "values" method".

        Returns
        -------
        sc: matplotlib.axes
            The axes are returned.

        """
        if models is None:
            models = values.index
            models = self.mls.oseries.loc[models, "z"].sort_values(
                ascending=False).index
        else:
            values = values.loc[models]

        x = self.mls.oseries.loc[models, "x"].astype(float)
        y = self.mls.oseries.loc[models, "y"].astype(float)
        s = self._normalize(self.mls.oseries.loc[models, "z"].astype(float), s)

        if show_nan:
            nan = values[values.isnull()].fillna(-999)
            plt.scatter(x[nan.index], y[nan.index], c="none", edgecolors="k",
                        s=s)

        sc = plt.scatter(x, y, c=values, s=s, edgecolors="k", **kwargs)

        if label:
            not_nan = values[~values.isnull()].index
            labels = values[not_nan].astype(str)
            for name, xy in zip(labels, zip(x[not_nan], y[not_nan])):
                plt.annotate(s=name, xy=xy,
                             bbox=dict(facecolor='w', edgecolor='k'),
                             textcoords="offset points", xytext=(10, 10))

        return sc

    def series(self, kind="stresses", label=False, **kwargs):
        """Plot the location of the oseries or the stresses on a map.

        Parameters
        ----------
        kind: str
            kind of series to plot. Possible values are the oseries,
            stresses or a specific type of stress (e.g. prec, evap or well).
        label: bool, optional
            Display a label next to the point with the name of the series.
        kwargs: dict, optional
            Any keyword arguments are passed on to plt.scatter.

        Returns
        -------
        sc: matplotlib.axes
            Return the axes.

        """
        if kind == "oseries":
            series = self.mls.oseries
        elif kind == "stresses":
            series = self.mls.stresses
        else:
            series = self.mls.stresses.loc[self.mls.stresses.kind == kind]

        x = series.loc[:, "x"].astype(float)
        y = series.loc[:, "y"].astype(float)

        sc = plt.scatter(x, y, **kwargs)

        if label:
            for name, xy in zip(x.index, zip(x, y)):
                plt.annotate(s=name, xy=xy, fontsize=10,
                             bbox=dict(facecolor='w', edgecolor='k'),
                             textcoords="offset points", xytext=(10, 10))

        return sc

    def _normalize(self, series, s=30):
        """Internal method to normalize the series for the size op the scatterplot.

        """
        mu = series.mean()
        if mu == 0.0:  # Prevent sizes of zero to be calculates
            mu = 1
        series = (series.subtract(series.min()) / mu + 1) * s
        return series



"""This files contains the Project class that deals with multiple models at
once.

Notes
-----
This module is created at Artesia Water by Raoul Collenteur.

Usage
-----

>>> mls = Project()

"""

from logging import getLogger
from os import getlogin

import numpy as np
import pandas as pd
from ..rfunc import Gamma
from ..timeseries import TimeSeries
from ..io.base import dump
from ..model import Model
from ..version import __version__
from ..stressmodels import StressModel2

from .maps import Map
from .plots import Plot

logger = getLogger(__name__)


class Project:
    """The Project class is a placeholder when multiple time series models
    are analyzed in a batch.

    """

    def __init__(self, name, metadata=None):
        """Initialize a Project instance.

        Parameters
        ----------
        name: str
            Name of the project
        metadata: dict
            Dictionary with any metadata information on the project.

        """
        self.models = {}
        self.name = name
        # Store the data in Pandas dataframes
        self.data = pd.DataFrame()

        # DataFrames to store the data of the oseries and stresses
        self.stresses = pd.DataFrame(columns=["name", "series", "kind", "x",
                                              "y", "z"])
        self.oseries = pd.DataFrame(columns=["name", "series", "kind", "x",
                                             "y", "z"])
        self.oseries.index.astype(str)

        # Project metadata and file information
        self.metadata = self.get_metadata(metadata)
        self.file_info = self._get_file_info()

        # Load other modules
        self.plots = Plot(self)
        self.maps = Map(self)

    def add_series(self, series, name=None, kind=None, metadata=None,
                   settings=None, **kwargs):
        """Method to add series to the oseries or stresses database.

        Parameters
        ----------
        series: pandas.Series / pastas.TimeSeries
            Series object.
        name: str
            String with the name of the series that will be maintained in
            the database.
        kind: str
            The kind of series that is added. When oseries are added it is
            necessary to state "oseries" here.
        metadata: dict
            Dictionary with any metadata that will be passed to the
            TimeSeries object that is created internally.
        settings: dict or str
            Dictionary with any settings that will be passed to the
            TimeSeries object that is created internally.

        Returns
        -------

        """
        if name is None:
            name = series.name

        if not isinstance(name, str):
            name = str(name)

        series.name = name

        if kind == "oseries":
            data = self.oseries
            if settings is None:
                settings = 'oseries'
        else:
            data = self.stresses

        if name in data.index:
            warning = ("Time series with name {} is already present in the "
                       "database. Existing series is overwitten.").format(name)
            logger.warning(warning)
        try:
            ts = TimeSeries(series=series, name=name, settings=settings,
                            metadata=metadata, **kwargs)
        except:
            logger.error("An error occurred. Time series {} is omitted "
                         "from the database.".format(name))
            return

        data.at[name, "name"] = name
        data.at[name, "series"] = ts  # Do not add as first!
        data.at[name, "kind"] = kind

        # Transfer the metadata (x, y and z) to dataframe as well to increase speed.
        for i in ts.metadata.keys():
            value = ts.metadata[i]
            data.at[name, i] = value

    def del_oseries(self, name):
        """Method that savely removes oseries from the project. It validates
        that the oseries is not used in any model.

        Parameters
        ----------
        name: str
            string with a single oseries name.

        """
        if name not in self.oseries.index:
            error = ("Time series with name {} is not present in the database."
                     " Please provide a different name.").format(name)
            logger.error(error)
        else:
            self.oseries.drop(name, inplace=True)
            logger.info(
                "Oseries with name {} deleted from the database.".format(name))

    def del_stress(self, name):
        """Method that removes oseries from the project.

        Parameters
        ----------
        stress: list or str
            list with multiple or string with a single oseries name.

        Returns
        -------

        """
        if name not in self.stresses.index:
            error = ("Stress with name {} is not present in the database."
                     " Please provide a different name.").format(name)
            logger.error(error)
        else:
            self.stresses.drop(name, inplace=True)
            logger.info(
                "Stress with name {} deleted from the database.".format(name))

    def add_model(self, oseries, model_name=None, **kwargs):
        """Method to add a Pastas Model instance based on one of the oseries.

        Parameters
        ----------
        oseries: str
            string with the exact names of one of the oseries indices.
        model_name: str
            Name of the model
        kwargs: dict
            any arguments that are taken by the Pastas Model instance can be
            provided.

        Returns
        -------
        ml: pastas.Model
            Pastas Model generated with the oseries and arguments provided.

        """
        if model_name is None:
            model_name = oseries

        # Validate name and ml_name before continuing
        if model_name in self.models.keys():
            warning = ("Model name {} is not unique, existing model is "
                       "overwritten.").format(model_name)
            logger.warning(warning)
        if oseries not in self.oseries.index:
            error = ("Oseries name {} is not present in the database. Make "
                     "sure to provide a valid name.").format(model_name)
            logger.error(error)
            return

        oseries = self.oseries.loc[oseries, "series"]
        ml = Model(oseries, name=model_name, **kwargs)

        # Add new model to the models dictionary
        self.models[model_name] = ml

        return ml

    def del_model(self, ml_name):
        """Method to safe-delete a model from the project.

        Parameters
        ----------
        model_name: str
            String with the model name.

        """
        name = self.models.pop(ml_name, None)
        info = "Model with name {} deleted from the database.".format(name)
        logger.info(info)

    def update_model_series(self):
        """Update all the Model series by their originals in self.oseries and self.stresses.
        This can for example be usefull when new data is added to any of the series in pr.oseries and pr.stresses

        """
        for name in self.models:
            ml = self.models[name]
            ml.oseries.series_original = self.oseries.loc[
                name, 'series'].series_original
            for sm in ml.stressmodels:
                for sm in ml.stressmodels:
                    for st in ml.stressmodels[sm].stress:
                        st.series_original = self.stresses.loc[
                            st.name, 'series'].series_original
            # set oseries_calib empty, so it is determined again the next time
            ml.oseries_calib = None

    def add_recharge(self, ml, rfunc=Gamma, name="recharge", **kwargs):
        """Adds a recharge element to the time series model. The
        selection of the precipitation and evaporation time series is based
        on the shortest distance to the a stresses in the stresseslist.

        Returns
        -------

        """
        key = str(ml.oseries.name)
        prec_name = self.get_nearest_stresses(key, kind="prec").iloc[0][0]
        prec = self.stresses.loc[prec_name, "series"]
        evap_name = self.get_nearest_stresses(key, kind="evap").iloc[0][0]
        evap = self.stresses.loc[evap_name, "series"]

        recharge = StressModel2([prec, evap], rfunc, name=name, **kwargs)

        ml.add_stressmodel(recharge)

    def get_nearest_stresses(self, oseries=None, stresses=None, kind=None,
                             n=1):
        """Method to obtain the nearest (n) stresses of a specific kind.

        Parameters
        ----------
        oseries: str
            String with the name of the oseries
        kind:
            String with the name of the stresses
        n: int
            Number of stresses to obtain

        Returns
        -------
        stresses:
            List with the names of the stresses.

        """

        distances = self.get_distances(oseries, stresses, kind)

        sorted = pd.DataFrame(columns=np.arange(n))

        for series in distances.index:
            series = pd.Series(distances.loc[series].sort_values().index[:n],
                               name=series)
            sorted = sorted.append(series)

        return sorted

    def get_distances(self, oseries=None, stresses=None, kind=None, ):
        """Method to obtain the distances in meters between the stresses and
        oseries.

        Parameters
        ----------
        oseries: str or list of str
        stresses: str or list of str
        kind: str

        Returns
        -------
        distances: pandas.DataFrame
            Pandas DataFrame with the distances between the oseries (index)
            and the stresses (columns).

        """
        if isinstance(oseries, str):
            oseries = [oseries]
        elif oseries is None:
            oseries = self.oseries.index

        if stresses is None and kind is None:
            stresses = self.stresses.index
        elif stresses is None:
            stresses = self.stresses[self.stresses.kind == kind].index
        elif stresses is not None and kind is not None:
            mask = self.stresses.kind == kind
            stresses = self.stresses.loc[stresses].loc[mask].index

        xo = pd.to_numeric(self.oseries.loc[oseries, "x"])
        xt = pd.to_numeric(self.stresses.loc[stresses, "x"])
        yo = pd.to_numeric(self.oseries.loc[oseries, "y"])
        yt = pd.to_numeric(self.stresses.loc[stresses, "y"])

        xh, xi = np.meshgrid(xt, xo)
        yh, yi = np.meshgrid(yt, yo)

        distances = pd.DataFrame(np.sqrt((xh - xi) ** 2 + (yh - yi) ** 2),
                                 index=oseries, columns=stresses)

        return distances

    def get_parameters(self, parameters, models=None, param_value="optimal"):
        """Method to get the parameters from each model. NaN-values are
        returned when the parameters is not present in the model.

        Parameters
        ----------
        parameters: list
            List with the names of the parameters. The parameter does not
            have to be used in all models.
        models: list
            List with the names of the models. These have to be in the
            Project models dictionary.
        param_value: str
            String with the parameter value that needs to be collected:
            Options are: initial, optimal (default), pmax, pmin and vary.

        Returns
        -------
        data: pandas.DataFrame or pandas.Series
            Returns a pandas DataFrame with the models name as the index and
            the parameters as columns. A pandas Series is returned when only
            one parameter values is collected.

        """
        if models is None:
            models = self.models.keys()

        data = pd.DataFrame(index=models, columns=parameters)

        for ml_name in models:
            ml = self.models[ml_name]
            for parameter in parameters:
                if parameter in ml.parameters.index:
                    value = ml.parameters.loc[parameter, param_value]
                    data.loc[ml_name, parameter] = value

        data = data.squeeze()
        return data.astype(float)

    def get_statistics(self, statistics, models=None, **kwargs):
        """Method to get the statistics for each model.

        Parameters
        ----------
        statistics: list
            List with the names of the statistics to calculate for each model.
        models: list
            List with the names of the models. These have to be in the
            Project models dictionary.

        Returns
        -------
        data: pandas.DataFrame or pandas.Series

        """
        if models is None:
            models = self.models.keys()

        data = pd.DataFrame(index=models, columns=statistics)

        for ml_name in models:
            ml = self.models[ml_name]
            for statistic in statistics:
                value = ml.stats.__getattribute__(statistic)(**kwargs)
                data.loc[ml_name, statistic] = value

        data = data.squeeze()
        return data.astype(float)

    def get_locations_geodataframe(self, models=None, **kwargs):
        import geopandas as gpd
        from shapely.geometry import Point

        if models is None:
            models = self.models.keys()

        data = pd.DataFrame(index=models)

        data = data.join(self.oseries.loc[models, ["x", "y", "z"]])
        data["geometry"] = [Point(xy[0], xy[1]) for xy in
                            self.oseries.loc[models, ["x", "y"]].values]
        data = gpd.GeoDataFrame(data, geometry="geometry", **kwargs)
        return data

    def get_oseries_metadata(self, oseries, metadata):
        """

        Parameters
        ----------
        oseries
        metadata

        Returns
        -------
        data: pandas.DataFrame

        """
        data = pd.DataFrame(data=None, index=oseries, columns=metadata)

        for oseries in data.index:
            meta = self.oseries.loc[oseries, "series"].metadata
            for key in metadata:
                data.loc[oseries, key] = meta[key]

        return data

    def get_oseries_settings(self, oseries, settings):
        """Method to obtain the settings from each oseries TimeSeries object.

        Parameters
        ----------
        oseries
        settings

        Returns
        -------
        data: pandas.DataFrame
            Pandas DataFrame with the oseries as index, settings as columns
            and the values as data.

        """
        data = pd.DataFrame(data=None, index=oseries, columns=settings)

        for oseries in data.index:
            sets = self.oseries.loc[oseries, "series"].settings
            for key in settings:
                data.loc[oseries, key] = sets[key]

        return data

    def get_metadata(self, meta=None):
        metadata = dict(
            projection=None
        )
        if meta:
            metadata.update(meta)

        return metadata

    def _get_file_info(self):
        file_info = dict()
        file_info["date_created"] = pd.Timestamp.now()
        file_info["date_modified"] = pd.Timestamp.now()
        file_info["pastas_version"] = __version__
        try:
            file_info["owner"] = getlogin()
        except:
            file_info["owner"] = "Unknown"
        return file_info

    def dump(self, fname=None, **kwargs):
        """Method to write a Pastas project to a file.

        Parameters
        ----------
        fname

        Returns
        -------

        """
        data = self.dump_data(**kwargs)
        return dump(fname, data)

    def dump_data(self, series=False, sim_series=False):
        """Method to export a Pastas Project and return a dictionary with
        the data to be exported.

        Parameters
        ----------
        series: bool
            export model input-series when True. Only export the name of
            the model input_series when False
            
        sim_series: bool
            export model output-series when True

        Returns
        -------
        data: dict
            A dictionary with all the project data

        """
        data = dict(
            name=self.name,
            models=dict(),
            metadata=self.metadata,
            file_info=self.file_info
        )

        # Series DataFrame
        data["oseries"] = self._series_to_dict(self.oseries)
        data["stresses"] = self._series_to_dict(self.stresses)

        # Models
        data["models"] = dict()
        for name, ml in self.models.items():
            data["models"][name] = ml.dump_data(series=series,
                                                sim_series=sim_series,
                                                file_info=False)

        return data

    def _series_to_dict(self, series):
        series = series.to_dict(orient="index")

        for name in series.keys():
            ts = series[name]["series"]
            series[name]["series"] = ts.dump(series=True)

        return series



"""Statistics for the Pastas Model class.

Statistics can be calculated for the following time series:
- Observation series
- Simulated series
- Residual series
- Noise series
Each of these series can be obtained through their individual (private) get
method for a specific time frame.
two different types of statistics are provided: model statistics and
descriptive statistics for each series.

Examples
--------

    >>> ml.stats.summary()
                                         Value
    Statistic
    Pearson R^2                       0.874113
    Root mean squared error           0.432442
    Bayesian Information Criterion  113.809120
    Average Deviation                 0.335966
    Explained variance percentage    72.701968
    Akaike InformationCriterion      25.327385

"""

from numpy import sqrt, log, nan
from pandas import DataFrame

from .decorators import model_tmin_tmax


class Statistics:
    # Save all statistics that can be calculated.
    ops = {'evp': 'Explained variance percentage',
           'rmse': 'Root mean squared error',
           'rmsi': 'Root mean squared noise',
           'sse': 'Sum of squares of the error',
           'avg_dev': 'Average Deviation',
           'rsq': 'Pearson R^2',
           'rsq_adj': 'Adjusted Pearson R^2',
           'bic': 'Bayesian Information Criterion',
           'aic': 'Akaike Information Criterion',
           'nse': 'Nash-Sutcliffe Efficiency'}

    def __init__(self, ml):
        """
        To obtain a list of all statistics that are
        included type:

        >>> print(ml.stats.ops)

        ml: Pastas Model
            ml is a time series Model that is calibrated.
        """
        # Save a reference to the model.
        self.ml = ml

    def __repr__(self):
        msg = """This module contains all the statistical functions that are
included in Pastas. To obtain a list of all statistics that are included type:

    >>> print(ml.stats.ops)"""
        return msg

    # The statistical functions
    @model_tmin_tmax
    def rmse(self, tmin=None, tmax=None):
        """Root mean squared error of the residuals.

        Notes
        -----
        .. math:: rmse = sqrt(sum(residuals**2) / N)

        where N is the number of residuals.
        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax).values
        N = res.size
        return sqrt((res ** 2).sum() / N)

    @model_tmin_tmax
    def rmsi(self, tmin=None, tmax=None):
        """Root mean squared error of the noise.

        Returns
        -------
        float or nan
            Return a float if noisemodel is present, nan if not.

        Notes
        -----
        .. math:: rmsi = sqrt(sum(noise**2) / N)

        where N is the number of noise.
        """
        if not self.ml.noisemodel:
            return nan
        else:
            res = self.ml.noise(tmin=tmin, tmax=tmax).values
            N = res.size
            return sqrt((res ** 2).sum() / N)

    @model_tmin_tmax
    def sse(self, tmin=None, tmax=None):
        """Sum of the squares of the error (SSE)

        Notes
        -----
        The SSE is calculated as follows:

        .. math:: SSE = sum(E ** 2)

        Where E is an array of the residual series.

        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax).values
        return (res ** 2).sum()

    @model_tmin_tmax
    def avg_dev(self, tmin=None, tmax=None):
        """Average deviation of the residuals.

        Notes
        -----
        .. math:: avg_dev = sum(E) / N

        Where N is the number of the residuals.

        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax).values
        return res.mean()

    @model_tmin_tmax
    def nse(self, tmin=None, tmax=None):
        """Nash-Sutcliffe coefficient for model fit.

        References
        ----------
        .. [NS] Nash, J. E., & Sutcliffe, J. V. (1970). River flow forecasting
        through conceptual models part I-A discussion of principles. Journal
        of hydrology, 10(3), 282-290.

        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax).values
        obs = self.ml.observations(tmin=tmin, tmax=tmax).values
        E = 1 - (res ** 2).sum() / ((obs - obs.mean()) ** 2).sum()
        return E

    @model_tmin_tmax
    def evp(self, tmin=None, tmax=None):
        """Explained variance percentage.

        Notes
        -----
        Commonly used statistic in time series models of groundwater levels.

        .. math:: evp = (var(h) - var(res)) / var(h) * 100%

        """
        res = self.ml.residuals(tmin=tmin, tmax=tmax).values
        obs = self.ml.observations(tmin=tmin, tmax=tmax).values
        if obs.var() == 0.0:
            return 100.
        else:
            evp = max(0.0, 100 * (1 - (res.var(ddof=0) / obs.var(ddof=0))))
        return evp

    @model_tmin_tmax
    def rsq(self, tmin=None, tmax=None):
        """Correlation between observed and simulated series.

        """
        obs = self.ml.observations(tmin=tmin, tmax=tmax).values
        res = self.ml.residuals(tmin=tmin, tmax=tmax).values
        RSS = (res ** 2.0).sum()
        TSS = ((obs - obs.mean()) ** 2.0).sum()
        return 1.0 - RSS / TSS

    @model_tmin_tmax
    def rsq_adj(self, tmin=None, tmax=None):
        """R-squared Adjusted for the number of free parameters.

        Notes
        -----
        .. math:: R_{corrected} = 1 - (n-1) / (n-N_param) * RSS/TSS

        Where:
            RSS = sum of the squared residuals.
            TSS = total sum of squared residuals
            N = Number of observations
            N_Param = Number of free parameters
        """

        obs = self.ml.observations(tmin=tmin, tmax=tmax).values
        res = self.ml.residuals(tmin=tmin, tmax=tmax).values
        N = obs.size
        RSS = (res ** 2.0).sum()
        TSS = ((obs - obs.mean()) ** 2.0).sum()
        nparam = self.ml.parameters.index.size
        return 1.0 - (N - 1.0) / (N - nparam) * RSS / TSS

    @model_tmin_tmax
    def bic(self, tmin=None, tmax=None):
        """Bayesian Information Criterium. The noise is used if a noisemodel is
         present, otherwise the residuals are used.

        Notes
        -----
        The Bayesian Information Criterium is calculated as follows:

        .. math:: BIC = -2 log(L) + nparam * log(N)

        Where:
            nparam : Number of free parameters



        """
        if self.ml.noisemodel:
            noise = self.ml.noise(tmin=tmin, tmax=tmax).values
        else:
            noise = self.ml.residuals(tmin=tmin, tmax=tmax).values
        n = noise.size
        nparam = self.ml.parameters[self.ml.parameters.vary == True].index.size
        bic = -2.0 * log((noise ** 2.0).sum()) + nparam * log(n)
        return bic

    @model_tmin_tmax
    def aic(self, tmin=None, tmax=None):
        """Akaike Information Criterium (AIC). The noise is used if a
        noisemodel is present, otherwise the residuals are used.

        Notes
        -----
        .. math:: AIC = -2 log(L) + 2 nparam

        Where
            nparam = Number of free parameters
            L = likelihood function for the model.

        """
        if self.ml.noisemodel:
            noise = self.ml.noise(tmin=tmin, tmax=tmax).values
        else:
            noise = self.ml.residuals(tmin=tmin, tmax=tmax).values
        nparam = self.ml.parameters[self.ml.parameters.vary == True].index.size
        aic = -2.0 * log((noise ** 2.0).sum()) + 2.0 * nparam
        return aic

    @model_tmin_tmax
    def summary(self, tmin=None, tmax=None, stats='basic'):
        """Prints a summary table of the model statistics. The set of
        statistics that are printed are stats by a dictionary of the desired
        statistics.

        Parameters
        ----------
        tmin
        tmax
        stats : str or dict
            dictionary of the desired statistics or a string with one of the
            predefined sets. Supported options are: 'basic', 'all', and 'dutch'

        Returns
        -------
        stats : Pandas.DataFrame
            single-column DataFrame with calculated statistics

        """
        output = {
            'basic': {
                'evp': 'Explained variance percentage',
                'rmse': 'Root mean squared error',
                'avg_dev': 'Average Deviation',
                'rsq': 'Pearson R^2',
                'bic': 'Bayesian Information Criterion',
                'aic': 'Akaike Information Criterion'},
        }

        # get labels and method names for stats output
        if stats == 'all':
            # sort by key, label, method name
            selected_output = sorted([(k, l, f) for k, d in output.items()
                                      for f, l in d.items()])
        else:
            # sort by name, method name
            selected_output = sorted([(0, l, f) for f, l in
                                      output[stats].items()])

        # compute statistics
        labels_and_values = [(l, getattr(self, f)(tmin=tmin, tmax=tmax))
                             for _, l, f in selected_output]
        labels, values = zip(*labels_and_values)

        stats = DataFrame(index=list(labels), data=list(values),
                          columns=['Value'])
        stats.index.name = 'Statistic'
        return stats

    @model_tmin_tmax
    def many(self, tmin=None, tmax=None, stats=None):
        """This method returns the values for a provided list of statistics.

        Parameters
        ----------
        tmin
        tmax
        stats: list
            list of statistics that need to be calculated.

        Returns
        -------

        """
        if not stats:
            stats = ['evp', 'rmse', 'rmsi', 'rsq']

        data = DataFrame(index=[0], columns=stats)
        for k in stats:
            data.iloc[0][k] = (getattr(self, k)(tmin=tmin, tmax=tmax))

        return data

    @model_tmin_tmax
    def all(self, tmin=None, tmax=None):
        """Returns a dictionary with all the statistics.

        Parameters
        ----------
        tmin: str
        tmax: str

        Returns
        -------
        stats: pd.DataFrame
            Dataframe with all possible statistics

        """
        stats = DataFrame(columns=['Value'])
        for k in self.ops.keys():
            stats.loc[k] = (getattr(self, k)(tmin=tmin, tmax=tmax))

        return stats



"""recharge_func module

Author: R.A. Collenteur

Contains the classes for the different models that are available to calculate
the recharge from evaporation and precipitation data.

Each Recharge class contains at least the following:

Attributes
----------
nparam: int
    Number of parameters needed for this model.

Functions
---------
set_parameters(self, name)
    A function that returns a Pandas DataFrame of the parameters of the
    recharge function. Columns of the dataframe need to be ['value', 'pmin',
    'pmax', 'vary']. Rows of the DataFrame have names of the parameters. Input
    name is used as a prefix. This function is called by a Tseries object.
simulate(self, evap, prec, p=None)
    A function that returns an array of the simulated recharge series.

"""

import pandas as pd
import numpy as np
import numba


class Linear:
    """Linear recharge model

    The recharge to the groundwater is calculated as:
    R = P - f * E

    """

    def __init__(self):
        self.nparam = 1

    def set_parameters(self, name):
        parameters = pd.DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        parameters.loc[name + '_f'] = (-1.0, -5.0, 0.0, 1, name)
        return parameters

    def simulate(self, precip, evap, p=None):
        recharge = precip + p * evap
        return recharge


class Percolation:
    """
    Percolation flow recharge model

    Other water balance for the root zone s calculated as:

    dS/dt = Pe - Kp * (Sr / Srmax)**Gamma - Epu * min(1, Sr / (0.5 * Srmax))

    """

    def __init__(self):
        self.nparam = 4
        self.dt = 1
        self.solver = 1

    def set_parameters(self, name):
        parameters = pd.DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        parameters.loc[name + '_Srmax'] = (0.26, np.nan, np.nan, 0, name)
        parameters.loc[name + '_Kp'] = (1.0e-2, 0.0, np.nan, 1, name)
        parameters.loc[name + '_Gamma'] = (3.0, 0.0, np.nan, 1, name)
        parameters.loc[name + '_Imax'] = (1.5e-3, 0.0, np.nan, 0, name)
        return parameters

    def simulate(self, prec, evap, p=None):
        t = np.arange(len(prec))
        recharge = self.perc(t, prec.values, evap.values, p[0], p[1], p[2],
                             p[3], self.dt, self.solver)[0]
        return recharge

    @numba.jit
    def perc(self, t, P, E, Srmax=0.1, Kp=0.03, Gamma=2.0, Imax=0.001, dt=1.0,
             solver=0):
        n = int(len(t) / dt)
        error = 1.0e-5

        # Create an empty array to store the soil state in
        S = np.zeros(n)
        S[0] = 0.5 * Srmax  # Set the initial system state
        Si = np.zeros(n)
        Si[0] = 0.0
        Pe = np.zeros(n)
        Pe[0] = 0.0
        Ei = np.zeros(n)
        Ei[0] = 0.0
        Epu = np.zeros(n)
        Epu[0] = 0.0
        Ea = np.zeros(n)
        Ea[0] = 0.0

        for t in range(n - 1):
            # Fill intercEpution bucket with new rain
            Si[t + 1] = Si[t] + P[t + 1]
            # Calculate effective precipitation
            Pe[t + 1] = np.max([0.0, Si[t + 1] - Imax])
            Si[t + 1] = Si[t + 1] - Pe[t + 1]
            # Evaporation from intercEpution
            Ei[t + 1] = np.min([Si[t + 1], E[t + 1]])
            # Update intercEpution state
            Si[t + 1] = Si[t + 1] - Ei[t + 1]
            # Update potential evapotranspiration
            Epu[t + 1] = E[t + 1] - Ei[t + 1]
            Last_S = S[t]
            iteration = 0
            bisection = 1

            # Use explicit Euler scheme to find an initial estimate for the newton raphson-method
            S[t + 1] = np.max(
                [0.0, S[t] + dt * (Pe[t] - Kp * (S[t] / Srmax) ** Gamma -
                                   Epu[t] * np.min([1.0, (S[t] / (0.5 *
                                                                  Srmax))]))])

            if solver == 1:
                # Start the while loop for the newton-Raphson iteration
                while abs(Last_S - S[t + 1]) > error:
                    if iteration > 100:
                        break  # Check if the number of iterations is not too high
                    iteration += 1
                    Last_S = S[t + 1]

                    g = Last_S - S[t] - dt * (
                            Pe[t] - Kp * (Last_S / Srmax) ** Gamma -
                            Epu[t] * np.min([1, (Last_S / (0.5 *
                                                           Srmax))]))
                    # Derivative dEpuends on the state of the system
                    if Last_S > (0.5 * Srmax):
                        g_derivative = 1.0 - dt * (
                                -Gamma * Kp * (Last_S / Srmax) ** (Gamma - 1))
                    else:
                        g_derivative = 1.0 - dt * (
                                -Gamma * Kp * (Last_S / Srmax) ** (Gamma - 1) -
                                Epu[
                                    t] * (
                                        0.5 * Srmax))

                    # Check if there is no zero-division error
                    if np.isnan(g / g_derivative):
                        bisection = 0
                        break
                    # if there is no zero-division error
                    else:  # use newton raphson
                        S[t + 1] = Last_S - g / g_derivative

                if bisection == 0:
                    iteration = 0
                    a = S[t]
                    b = S[t + 1]
                    c = a + b / 2.0

                    while ((b - a) / 2.0) > error:
                        if iteration > 100:
                            print('iteration in bisection method exceeded 100',
                                  iteration)
                            break
                        iteration += 1  # increase the number of iterations by 1

                        if (c - S[t] - dt * (
                                Pe[t] - Kp * (c / Srmax) ** Gamma - Epu[t]
                                * np.min([1, (c / (0.5 * Srmax))]))) == 0.0:
                            return c  # Return the current value if it is correct
                        elif (a - S[t] - dt * (
                                Pe[t] - Kp * (a / Srmax) ** Gamma - Epu[
                            t] * np.min([1.0, (a / (0.5 * Srmax))]))) * (
                                c - S[t] - dt
                                * (Pe[t] - Kp * (c / Srmax) ** Gamma - Epu[
                            t] * np.min(
                            [
                                1.0, (c / (0.5 * Srmax))]))) > 0.0:
                            b = c
                        else:
                            a = c

                        c = a + b / 2.0

                    S[t + 1] = c

                assert ~np.isnan(
                    S[t + 1]), 'NaN-value calculated for soil state'

            # Make sure the solution is larger then 0.0 and smaller than Srmax
            S[t + 1] = np.min([Srmax, np.max([0.0, S[t + 1]])])
            Ea[t + 1] = Epu[t + 1] * np.min([1, (S[t + 1] / (0.5 * Srmax))])

        R = np.append(0.0, Kp * dt * 0.5 * (
                (S[:-1] ** Gamma + S[1:] ** Gamma) / (Srmax ** Gamma)))

        return R, S, Ea, Ei



"""The stressmodels module contains all the transforms that can be added to the
 simulation of a model. These transforms are applied after the simulation,
 to incorporate nonlineair effects.

"""
import numpy as np
from pandas import DataFrame


class ThresholdTransform:
    """ThresholdTransform lowers the simulation when it exceeds a certain value

    In geohydrology this transform can for example be used in a situation where
    the groundwater level reaches the surface level and forms a lake. Beacuase
    of the larger storage of the lake, the (groundwater) level then rises
    slower when it rains.


    Parameters
    ----------
    value : float
        The starting value above which the simulation is lowered
    vmin : float
        The minimum value above which the simulation is lowered
    vmin : float
        The maximum value above which the simulation is lowered
    name: str
        Name of the transform
    nparam : int
        The number of parameters. Default is nparam=2. The first parameter
        then is the threshold, and the second parameter is the factor with
        which the simulation is lowered.

    """
    _name = "ThresholdTransform"

    def __init__(self, value=np.nan, vmin=np.nan, vmax=np.nan,
                 name='ThresholdTransform', nparam=2):
        self.value = value
        self.vmin = vmin
        self.vmax = vmax
        self.name = name
        self.nparam = nparam

    def set_model(self, ml):
        obs = ml.observations()
        if np.isnan(self.value):
            self.value = obs.min() + 0.75 * (obs.max() - obs.min())
        if np.isnan(self.vmin):
            self.vmin = obs.min() + 0.5 * (obs.max() - obs.min())
        if np.isnan(self.vmax):
            self.vmax = obs.max()
        self.set_init_parameters()

    def set_init_parameters(self):
        self.parameters = DataFrame(
            columns=['initial', 'pmin', 'pmax', 'vary', 'name'])
        self.parameters.loc[self.name + '_1'] = (
            self.value, self.vmin, self.vmax, 1, self.name)
        if self.nparam == 2:
            self.parameters.loc[self.name + '_2'] = (0.5, 0., 1., 1, self.name)

    def simulate(self, h, p):
        if self.nparam == 1:
            # value above a threshold p[0] are equal to the threshold
            h[h > p[0]] = p[0]
        elif self.nparam == 2:
            # values above a threshold p[0] are scaled by p[1]
            mask = h > p[0]
            h[mask] = p[0] + p[1] * (h[mask] - p[0])
        else:
            raise ValueError('Not yet implemented yet')
        return h

    def dump(self):
        data = dict()
        data["transform"] = self._name
        data["value"] = self.value
        data["vmin"] = self.vmin
        data["vmax"] = self.vmax
        data["name"] = self.name
        data['nparam'] = self.nparam
        return data



"""
This python script can be used to solve the differential equations that are used
for the root zone module. The output of the function is the soil state (how
saturated it is) at each timestep, and the groundwater recharge N.

A .pyx-file is available of this script that can be cythonized and compiled, increasing
computation speeds up to 35 times. The use of this compiled version is strongly
recommended. Take a look at cythonize.py for compilation instructions.

Three models can be used:
-------------------------
- Percolation Flow:
dS/dt = Pe[t] - Kp * (Sr/Srmax)**Gamma - Epu * min(1, Sr/0.5Srmax) 

- Preferential Flow:
dS/ Dt = Pe[t] * (1 - (Sr[t] / Srmax)**Beta)- Epu * min(1, Sr/0.5Srmax)

- Combination:
dS/ Dt = Pe[t] * (1 - (Sr[t] / Srmax)**Beta) - Kp * (Sr/Srmax)**Gamma - Epu
         * min(1, Sr/0.5*Srmax)

Numerical info:
---------------
The soil module can be solved with an implicit or explicit euler scheme. Newton
Raphson iteration is used as the root finder, but it is switched to a bisection
method if this fails. The initial estimate for the the NR-iteration is provided
by an Explicit Euler solution of the above differential equation.

To Do:
------
- Built in more external / internal checks for water balance

References:
-----------
- R.A. Collenteur [2016] Non-linear time series analysis of deep groundwater
levels: Application to the Veluwe. MSc. thesis, TU Delft.
http://repository.tudelft.nl/view/ir/uuid:baf4fc8c-6311-407c-b01f-c80a96ecd584/

@author: Raoul Collenteur
"""

from __future__ import print_function, division

import numpy as np


def pref(t, P, E, Srmax=0.1, Beta=2.0, Imax=0.001, dt=1.0, solver=1):
    """
    In this section the preferential flow model is defined.
    dS/ Dt = Pe[t] * (1 - (Sr[t] / Srmax)**Beta)- Epu * min(1, Sr/0.5Srmax)
    """

    n = int(len(t) / dt)
    error = 1.0e-5

    # Create an empty array to store the soil state in
    S = np.zeros(n)
    S[0] = 0.5 * Srmax  # Set the initial system state
    Si = np.zeros(n)
    Si[0] = 0.0
    Pe = np.zeros(n)
    Pe[0] = 0.0
    Ei = np.zeros(n)
    Ei[0] = 0.0
    Epu = np.zeros(n)
    Epu[0] = 0.0
    Ea = np.zeros(n)
    Ea[0] = 0.0

    for t in range(n - 1):
        Si[t + 1] = Si[t] + P[t + 1]  # Fill intercEpution bucket with new rain
        Pe[t + 1] = np.max(
            [0.0, Si[t + 1] - Imax])  # Calculate effective precipitation
        Si[t + 1] = Si[t + 1] - Pe[t + 1]
        Ei[t + 1] = np.min(
            [Si[t + 1], E[t + 1]])  # Evaporation from intercEpution
        Si[t + 1] = Si[t + 1] - Ei[t + 1]  # Update intercEpution state
        Epu[t + 1] = E[t + 1] - Ei[
            t + 1]  # Update potential evapotranspiration

        Last_S = S[t]
        iteration = 0
        bisection = 1

        # Use explicit Euler scheme to find an initial estimate for the newton raphson-method

        S[t + 1] = np.max([0.0, S[t] + dt * (
            Pe[t] * (1 - (S[t] / Srmax) ** Beta) - Epu[t] * np.min([1.0, (
                S[t] / (0.5 * Srmax))]))])

        if solver == 1:  # If implicit euler is used
            # Start the while loop for the newton-Raphson iteration
            while abs(Last_S - S[t + 1]) > error:
                if iteration > 100:
                    break  # Check if the number of iterations is not too high
                iteration += 1
                Last_S = S[t + 1]

                g = Last_S - S[t] - dt * (
                    Pe[t] * (1 - (Last_S / Srmax) ** Beta) -
                    Epu[t] * np.min([1.0, (Last_S / (0.5 *
                                                     Srmax))]))
                # Derivative dEpuends on the state of the system
                if Last_S > (0.5 * Srmax):
                    g_derivative = 1.0 - dt * (
                        -Beta * Pe[t] * (Last_S / Srmax) **
                        (Beta - 1))
                else:
                    g_derivative = 1.0 - dt * (-Beta * Pe[t] * (Last_S / Srmax)
                                               ** (Beta - 1) - Epu[t] * (
                                                   0.5 * Srmax))

                # Check if there is no zero-division error
                if np.isnan(g / g_derivative):
                    bisection = 0
                    break
                # if there is no zero-division error                
                else:  # use newton raphson
                    S[t + 1] = Last_S - g / g_derivative

            if bisection == 0:
                iteration = 0
                a = S[t]
                b = S[t + 1]
                c = a + b / 2.0

                while ((b - a) / 2.0) > error:
                    if iteration > 100:
                        print('iteration in bisection method exceeded 100',
                              iteration)
                        break
                    iteration += 1  # increase the number of iterations by 1

                    if (c - S[t] - dt * (
                                    Pe[t] * (1 - (c / Srmax) ** Beta) - Epu[
                                t] *
                                np.min([1.0, (c / (0.5 * Srmax))]))) == 0.0:
                        return c  # Return the current value if it is correct
                    elif (a - S[t] - dt * (
                                    Pe[t] * (1 - (a / Srmax) ** Beta) - Epu[t]
                                * np.min([1.0, (a / (0.5 * Srmax))]))) * (
                                    c - S[t] - dt * (
                                            Pe[t] * (1 - (c / Srmax) ** Beta) -
                                            Epu[
                                                t] *
                                            np.min([1.0, (
                                                        c / (
                                                        0.5 * Srmax))]))) > 0.0:
                        b = c
                    else:
                        a = c

                    c = a + b / 2.0

                S[t + 1] = c

            assert ~np.isnan(S[t + 1]), 'NaN value calculated for soil state'

        S[t + 1] = np.min([Srmax,
                           np.max(
                               [0.0, S[t + 1]])])  # Make sure the solution is
        # larger
        #  then 0.0 and smaller than Srmax
        Ea[t + 1] = Epu[t + 1] * np.min([1.0, (S[t + 1] / (0.5 * Srmax))])

    R = np.append(0.0, Pe[1:] * dt * 0.5 * (
        (S[:-1] ** Beta + S[1:] ** Beta) / (Srmax ** Beta)))

    return R, S, Ea, Ei


""" -----------------------------------------------
In this section the percolation model is defined. 
dS/dt = Pe[t] - Kp * (Sr/Srmax)**Gamma - Epu * min(1, Sr/0.5Srmax) 
----------------------------------------------- """


def perc(t, P, E, Srmax=0.1, Kp=0.03, Gamma=2.0, Imax=0.001, dt=1.0, solver=1):
    n = int(len(t) / dt)
    error = 1.0e-5

    # Create an empty array to store the soil state in
    S = np.zeros(n)
    S[0] = 0.5 * Srmax  # Set the initial system state
    Si = np.zeros(n)
    Si[0] = 0.0
    Pe = np.zeros(n)
    Pe[0] = 0.0
    Ei = np.zeros(n)
    Ei[0] = 0.0
    Epu = np.zeros(n)
    Epu[0] = 0.0
    Ea = np.zeros(n)
    Ea[0] = 0.0

    for t in range(n - 1):
        Si[t + 1] = Si[t] + P[t + 1]  # Fill intercEpution bucket with new rain
        Pe[t + 1] = np.max([0.0, Si[t + 1] - Imax])  # Calculate effective
        # precipitation
        Si[t + 1] = Si[t + 1] - Pe[t + 1]
        Ei[t + 1] = np.min(
            [Si[t + 1], E[t + 1]])  # Evaporation from intercEpution
        Si[t + 1] = Si[t + 1] - Ei[t + 1]  # Update intercEpution state
        Epu[t + 1] = E[t + 1] - Ei[
            t + 1]  # Update potential evapotranspiration

        Last_S = S[t]
        iteration = 0
        bisection = 1
        # Use explicit Euler scheme to find an initial estimate for the newton raphson-method

        S[t + 1] = np.max(
            [0.0, S[t] + dt * (Pe[t] - Kp * (S[t] / Srmax) ** Gamma -
                               Epu[t] * np.min([1.0, (S[t] / (0.5 *
                                                              Srmax))]))])

        if solver == 1:
            # Start the while loop for the newton-Raphson iteration
            while abs(Last_S - S[t + 1]) > error:
                if iteration > 100:
                    break  # Check if the number of iterations is not too high
                iteration += 1
                Last_S = S[t + 1]

                g = Last_S - S[t] - dt * (
                    Pe[t] - Kp * (Last_S / Srmax) ** Gamma -
                    Epu[t] * np.min([1, (Last_S / (0.5 *
                                                   Srmax))]))
                # Derivative dEpuends on the state of the system
                if Last_S > (0.5 * Srmax):
                    g_derivative = 1.0 - dt * (
                        -Gamma * Kp * (Last_S / Srmax) ** (Gamma - 1))
                else:
                    g_derivative = 1.0 - dt * (
                        -Gamma * Kp * (Last_S / Srmax) ** (Gamma - 1) - Epu[
                            t] * (
                            0.5 * Srmax))

                # Check if there is no zero-division error            
                if np.isnan(g / g_derivative):
                    bisection = 0
                    break
                # if there is no zero-division error
                else:  # use newton raphson
                    S[t + 1] = Last_S - g / g_derivative

            if bisection == 0:
                iteration = 0
                a = S[t]
                b = S[t + 1]
                c = a + b / 2.0

                while ((b - a) / 2.0) > error:
                    if iteration > 100:
                        print('iteration in bisection method exceeded 100',
                              iteration)
                        break
                    iteration += 1  # increase the number of iterations by 1

                    if (c - S[t] - dt * (
                                    Pe[t] - Kp * (c / Srmax) ** Gamma - Epu[t]
                                * np.min([1, (c / (0.5 * Srmax))]))) == 0.0:
                        return c  # Return the current value if it is correct
                    elif (a - S[t] - dt * (
                                    Pe[t] - Kp * (a / Srmax) ** Gamma - Epu[
                                t] * np.min([1.0, (a / (0.5 * Srmax))]))) * (
                                    c - S[t] - dt
                                * (Pe[t] - Kp * (c / Srmax) ** Gamma - Epu[
                                    t] * np.min(
                                    [
                                        1.0, (c / (0.5 * Srmax))]))) > 0.0:
                        b = c
                    else:
                        a = c

                    c = a + b / 2.0

                S[t + 1] = c

            assert ~np.isnan(S[t + 1]), 'NaN-value calculated for soil state'

        S[t + 1] = np.min([Srmax, np.max([0.0, S[t + 1]])])  # Make sure the solution is
        # larger
        #  then 0.0 and smaller than Srmax
        Ea[t + 1] = Epu[t + 1] * np.min([1, (S[t + 1] / (0.5 * Srmax))])

    R = np.append(0.0, Kp * dt * 0.5 * (
        (S[:-1] ** Gamma + S[1:] ** Gamma) / (Srmax ** Gamma)))

    return R, S, Ea, Ei


"""-----------------------------------------------
In this section a combination of the percolation and the preferential flow model
is applied:

dS/ Dt = Pe[t] * (1 - (Sr[t] / Srmax)**Beta) - Kp * (Sr/Srmax)**Gamma -
Epu * min(1, Sr/0.5*Srmax)

----------------------------------------------- """


def comb(t, P, E, Srmax=0.1, Kp=0.03, Beta=2.0, Gamma=2.0, Imax=0.001,
         dt=1.0, solver=1):
    n = int(len(t) / dt)
    error = 1.0e-5

    # Create an empty array to store the soil state in
    S = np.zeros(n)
    S[0] = 0.5 * Srmax  # Set the initial system state
    Si = np.zeros(n)
    Si[0] = 0.0
    Pe = np.zeros(n)
    Pe[0] = 0.0
    Ei = np.zeros(n)
    Ei[0] = 0.0
    Epu = np.zeros(n)
    Epu[0] = 0.0
    Ea = np.zeros(n)
    Ea[0] = 0.0

    for t in range(n - 1):
        Si[t + 1] = Si[t] + P[t + 1]  # Fill intercEpution bucket with new rain
        Pe[t + 1] = np.max([0.0, Si[t + 1] - Imax])  # Calculate effective
        # precipitation
        Si[t + 1] = Si[t + 1] - Pe[t + 1]
        Ei[t + 1] = np.min(
            [Si[t + 1], E[t + 1]])  # Evaporation from interception
        Si[t + 1] = Si[t + 1] - Ei[t + 1]  # Update interception state
        Epu[t + 1] = E[t + 1] - Ei[
            t + 1]  # Update potential evapotranspiration

        Last_S = S[t]
        iteration = 0
        bisection = 1

        # Use explicit Euler scheme to find an initial estimate for the newton raphson-method

        S[t + 1] = np.max([0.0, S[t] + dt * (
            Pe[t] * (1 - (S[t] / Srmax) ** Beta) - Kp * (
                S[t] / Srmax) ** Gamma -
            Epu[
                t] * np.min([1.0, (S[t] / (0.5 * Srmax))]))])

        if solver == 1:  # If implicit euler is used
            # Start the while loop for the newton-Raphson iteration
            while abs(Last_S - S[t + 1]) > error:
                if iteration > 100:
                    break  # Check if the number of iterations is not too high
                iteration += 1
                Last_S = S[t + 1]

                g = Last_S - S[t] - dt * (
                    Pe[t] * (1 - (Last_S / Srmax) ** Beta) - Kp * (
                        Last_S / Srmax) ** Gamma - Epu[t] * np.min([1.0, (
                        Last_S / (0.5 * Srmax))]))
                # Derivative depends on the state of the system
                if Last_S > (0.5 * Srmax):
                    g_derivative = 1.0 - dt * (
                        -Beta * Pe[t] * (Last_S / Srmax) ** (
                            Beta - 1) - Gamma * Kp * (
                            Last_S / Srmax) ** (Gamma - 1))
                else:
                    g_derivative = 1.0 - dt * (
                        -Beta * Pe[t] * (Last_S / Srmax) ** (
                            Beta - 1) - Gamma * Kp * (
                            Last_S / Srmax) ** (Gamma - 1) - Epu[t] * (
                            0.5 * Srmax))

                # Check if there is no zero-division error            
                if np.isnan(g / g_derivative):
                    bisection = 0
                    break
                # if there is no zero-division error
                else:  # use newton raphson
                    S[t + 1] = Last_S - g / g_derivative
                    #
            if bisection == 0:
                iteration = 0
                a = S[t]
                b = S[t + 1]
                c = a + b / 2.0
                #
                while ((b - a) / 2.0) > error:
                    if iteration > 100:
                        print('iteration in bisection method exceeded 100',
                              iteration)
                        break
                    iteration += 1  # increase the number of iterations by 1

                    if (c - S[t] - dt * (
                                        Pe[t] * (
                                        1 - (c / Srmax) ** Beta) - Kp * (
                                            c / Srmax) ** Gamma - Epu[
                                t] * np.min([1, (
                                        c / (0.5 * Srmax))]))) == 0.0:
                        return c  # Return the current value if it is correct
                    elif (a - S[t] - dt * (
                                        Pe[t] * (
                                                1 - (
                                                a / Srmax) ** Beta) - Kp * (
                                            a / Srmax) ** Gamma - Epu[
                                t] * np.min([
                                1.0, (
                                            a / (0.5 * Srmax))]))) * (
                                    c - S[t] - dt * (
                                                Pe[t] * (
                                                        1 - (
                                                                c / Srmax) ** Beta) - Kp * (
                                                    c / Srmax) ** Gamma - Epu[
                                        t] * np.min([
                                        1.0, (
                                                    c / (
                                                            0.5 * Srmax))]))) > 0.0:
                        b = c
                    else:
                        a = c

                    c = a + b / 2.0

                S[t + 1] = c

            assert ~np.isnan(S[t + 1]), 'NaN-value calculated for soil state'

        S[t + 1] = np.min([Srmax, np.max([0.0, S[
            t + 1]])])  # Make sure the solution is larger then 0.0 and smaller
        # than Srmax
        Ea[t + 1] = Epu[t + 1] * np.min([1, (S[t + 1] / (0.5 * Srmax))])

    Rs = np.append(0.0, Kp * dt * 0.5 * (
        (S[:-1] ** Gamma + S[1:] ** Gamma) / (Srmax ** Gamma)))  # Percolation
    Rf = np.append(0.0, Pe[1:] * dt * 0.5 * (
        (S[:-1] ** Beta + S[1:] ** Beta) / (Srmax ** Beta)))  # Preferential

    return Rs, Rf, S, Ea, Ei



"""
@author: ruben calje

"""

import pandas as pd
import warnings

from ..timeseries import TimeSeries


def read_knmi(fname, variables='RD'):
    """This method can be used to import KNMI data.

    Parameters
    ----------
    fname: str
        Filename and path to a Dino file.
    variables: str
        String with the variable name to extract.

    Returns
    -------
    ts: pastas.TimeSeries
        returns a Pastas TimeSeries object or a list of objects.

    """
    knmi = KnmiStation.fromfile(fname)
    if variables is None:
        variables = knmi.variables.keys()
    if type(variables) == str:
        variables = [variables]

    stn_codes = knmi.data['STN'].unique()

    ts = []
    for code in stn_codes:
        for variable in variables:
            if variable not in knmi.data.keys():
                raise (ValueError(
                    "variable %s is not in this dataset. Please use one of "
                    "the following keys: %s" % (variable, knmi.data.keys())))

            series = knmi.data.loc[knmi.data['STN'] == code, variable]
            # get rid of the hours when data is daily
            if pd.infer_freq(series.index) == 'D':
                series.index = series.index.normalize()

            metadata = {}
            if knmi.stations is not None and not knmi.stations.empty:
                station = knmi.stations.loc[code, :]
                metadata['x'] = station.LON_east
                metadata['y'] = station.LAT_north
                metadata['z'] = station.ALT_m
                metadata['projection'] = 'epsg:4326'
                stationname = station.NAME
            else:
                stationname = str(code)
            metadata['description'] = knmi.variables[variable]
            if variable == 'RD' or variable == 'RH':
                kind = 'prec'
            elif variable == 'EV24':
                kind = 'evap'
            else:
                kind = None
            ts.append(TimeSeries(series, name=variable + ' ' + stationname,
                                 metadata=metadata, settings=kind))
    if len(ts) == 1:
        ts = ts[0]
    return ts


class KnmiStation:
    """
    Reads daily meteorological data in a file from stations of the KNMI:
    knmi = KnmiStation.fromfile(filename)

    Data can be downloaded for the meteorological stations at:
    https://www.knmi.nl/nederland-nu/klimatologie/daggegevens
    or
    http://projects.knmi.nl/klimatologie/daggegevens/selectie.cgi

    For the rainfall stations data is available at:
    https://www.knmi.nl/nederland-nu/klimatologie/monv/reeksen

    Also, data from the meteorological stations can be downloaded directly, for example with
    knmi = KnmiStation(stns=260, start=datetime(1970, 1, 1), end=datetime(1971, 1, 1))  # 260 = de bilt
    knmi.download()

    Hourly data can be downloaded with the 'interval'keyword set to 'hour' or 'hourly':
    knmi = KnmiStation(stns=260, start='2017', end='2018', interval='hourly')

    Data from rainfall-stations can be downloaded by asking for the variable 'RD' (the stns variable now describes codes for rainfall-stations):
    knmi = KnmiStation(stns=550, start='2018', end='2019', vars='RD') # rainfall-station in de bilt

    Times are recalculated to UT+1 (standard-time in the Netherlands), from UT.
    Also the datetime-index of the data is set at the end of the period that the data describes.
    So the rainfall between 2018-01-01 09:00:00 (08:00:00 UT) and 2018-01-02 09:00:00 (08:00:00 UT) gets the timestamp of 2018-01-02 09:00:00

    Units in the data of the knmi are recalculated to more basic SI-units. So mm are transformed to m, and a factor of 0,1 is transformed to 1.

    A description of the variables is found in knmi.variables.
    Information about the measurement-station(s) is found in knmi.stations.
    The measurement-data itself is found in knmi.data
    """

    def __init__(self, *args, **kwargs):
        self.stations = pd.DataFrame()
        self.variables = dict()
        self.data = pd.DataFrame()
        if len(args) > 0 or len(kwargs) > 0:
            warnings.warn("In the future use KnmiStation.download(**kwargs) "
                          "instead "
                          "of KnmiStation(**kwargs)",
                          FutureWarning)
            self._download(*args, **kwargs)
            # diable download method, as old code will call this again
            self.download = lambda *args, **kwargs: None
        else:
            # change download method to the instance-method
            self.download = self._download

    # Construct KnmiStation from file
    @classmethod
    def fromfile(cls, fname):
        self = cls()
        with open(fname, 'r') as f:
            self.readdata(f)
        f.close()
        return self

    # Construct KnmiStation from download
    @classmethod
    def download(cls, start=None, end=None, inseason=False, vars='ALL',
                 stns=260, interval='daily'):
        """

        """
        self = cls()
        self._download(start=start, end=end, inseason=inseason, vars=vars,
                       stns=stns, interval=interval)
        return self

    def _download(self, start=None, end=None, inseason=False, vars='ALL',
                  stns=260, interval='daily'):
        # Import the necessary modules (optional and not included in the
        # installation of pastas).
        try:
            import requests
        except ImportError:
            raise ImportError(
                'The module requests could not be imported. '
                'Please install through:'
                '>>> pip install requests'
                'or:'
                '>>> conda install requests')

        from io import StringIO

        if start is None:
            start = pd.Timestamp(pd.Timestamp.today().year, 1, 1)
        else:
            start = pd.to_datetime(start)
        if end is None:
            end = pd.Timestamp.today()
        else:
            end = pd.to_datetime(end)

        if interval.startswith('hour') and vars == 'RD':
            raise (
                ValueError('Interval can not be hourly for rainfall-stations'))

        if interval.startswith('hour'):
            # hourly data from meteorological stations
            url = 'http://projects.knmi.nl/klimatologie/uurgegevens/getdata_uur.cgi'
        elif vars == 'RD':
            # daily data from rainfall-stations
            url = 'http://projects.knmi.nl/klimatologie/monv/reeksen/getdata_rr.cgi'
        else:
            # daily data from meteorological stations
            url = 'http://projects.knmi.nl/klimatologie/daggegevens/getdata_dag.cgi'
        if not isinstance(stns, str):
            if isinstance(stns, int):
                stns = str(stns)
            else:
                stns = [str(i) for i in stns]
                stns = ":".join(stns)

        if interval.startswith('hour'):
            data = {
                'start': start.strftime('%Y%m%d') + '01',
                'end': end.strftime('%Y%m%d') + '24',
                'vars': vars,
                'stns': stns,
            }
        else:
            data = {
                'start': start.strftime('%Y%m%d'),
                'end': end.strftime('%Y%m%d'),
                'inseason': str(int(inseason)),
                'vars': vars,
                'stns': stns,
            }
        result = requests.get(url, params=data).text

        f = StringIO(result)
        self.readdata(f)

    def readdata(self, f):
        stations = pd.DataFrame()
        variables = dict()

        isLocations = False
        line = f.readline()
        isMeteo = line.startswith('# ')

        # Process the header information (Everything < 'STN,')
        while 'STN,' not in line:
            # Pre-format the line
            line = line.strip('\n')
            line = line.lstrip('# ')

            # If line is empty, skipline
            if line.strip() == '':
                pass
            # If line contains station info (can only happen for meteorological stations)
            elif isMeteo and line.startswith('STN '):
                isLocations = True
                line = line.strip()
                titels = line.split()
                titels = [x.replace('(', '_') for x in titels]
                titels = [x.replace(r')', '') for x in titels]

                # Create pd.DataFrame for station data
                stations = pd.DataFrame(columns=titels)
                stations.set_index(['STN'], inplace=True)

            # If line contains variables
            elif ' = ' in line:
                isLocations = False
                varDes = line.split(' = ')
                variables[varDes[0].strip()] = varDes[1].strip()
            # If location data is recognized in the previous line
            elif isLocations:
                # Format line. Ensure delimiter is two spaces to read the
                # location correctly
                line = line.strip()
                line = line.replace(':', '')
                line = line.replace('         ', '  ')
                line = line.replace('        ', '  ')
                line = line.replace('       ', '  ')
                line = line.replace('      ', '  ')
                line = line.replace('     ', '  ')
                line = line.replace('    ', '  ')
                line = line.replace('   ', '  ')
                # Add station location data
                line = line.split('  ')
                stn = int(line[0])

                def maybe_float(s):
                    try:
                        return float(s)
                    except (ValueError, TypeError):
                        return s

                line = [maybe_float(v) for v in line[1:]]
                stations.loc[stn] = line

            # Read in a new line and start over
            line = f.readline()

        # The header information of the datablock
        line = line.strip('\n')
        line = line.lstrip('# ')
        header = line.split(',')
        header = [item.lstrip().rstrip() for item in header]
        pos = f.tell()
        line = f.readline()  # Skip empty line after header
        if line not in ["\n", "\r\n", "# \n", '# \r\n']:
            # sometimes there is no empty line between the header and the data
            f.seek(pos)
        # Process the datablock
        if False:
            # older method, is much slower
            string2datetime = lambda x: pd.to_datetime(x, format='%Y%m%d')

            data = pd.read_csv(f, header=None, names=header,
                               parse_dates=['YYYYMMDD'], index_col='YYYYMMDD',
                               na_values='     ',
                               converters={1: string2datetime})
        else:
            # newer method, calculating the date afterwards is much faster
            data = pd.read_csv(f, header=None, names=header, na_values='     ')
            data.set_index(pd.to_datetime(data.YYYYMMDD, format='%Y%m%d'),
                           inplace=True)
            data = data.drop('YYYYMMDD', axis=1)

        # convert the hours if provided
        if 'HH' in data.keys():
            # hourly data, Hourly division 05 runs from 04.00 UT to 5.00 UT
            data.index = data.index + pd.to_timedelta(data['HH'], unit='h')
            data.pop('HH')
        else:
            # daily data
            if 'RD' in data.keys():
                # daily precipitation amount in 0.1 mm over the period 08.00 preceding day - 08.00 UTC present day
                data.index = data.index + pd.to_timedelta(8, unit='h')
            else:
                # add a full day for meteorologiscal data, so that the timestamp is at the end of the period that the data represenets
                data.index = data.index + pd.to_timedelta(1, unit='d')

        # from UT to UT+1 (standard-time in the Netherlands)
        data.index = data.index + pd.to_timedelta(1, unit='h')

        # Delete empty columns
        if '' in data.columns:
            data.drop('', axis=1, inplace=True)

        # Adjust the unit of the measurements
        for key, value in variables.items():
            # test if key existst in data
            if key not in data.keys():
                if key == 'YYYYMMDD' or key == 'HH':
                    pass
                elif key == 'T10N':
                    variables.pop(key)
                    key = 'T10'
                else:
                    raise NameError(key + ' does not exist in data')
            if ' (-1 for <0.05 mm)' in value or ' (-1 voor <0.05 mm)' in value:
                # set 0.025 mm where data == -1
                data.loc[data[key] == -1, key] = 0.25  # unit is still 0.1 mm
                value = value.replace(' (-1 for <0.05 mm)', '')
                value = value.replace(' (-1 voor <0.05 mm)', '')
            if '0.1 ' in value:
                # transform 0.1 to 1
                data[key] = data[key] * 0.1
                value = value.replace('0.1 ', '')
            if ' tiende ' in value:
                # transform 0.1 to 1
                data[key] = data[key] * 0.1
                value = value.replace(' tiende ', ' ')
            if ' mm' in value:
                # transform mm to m
                data[key] = data[key] * 0.001
                value = value.replace(' mm', ' m')
            if ' millimeters' in value:
                # transform mm to m
                data[key] = data[key] * 0.001
                value = value.replace(' millimeters', ' m')
            if '(in percents)' in value:
                # do not adjust (yet)
                pass
            if 'hPa' in value:
                # do not adjust (yet)
                pass
            if 'J/cm2' in value:
                # do not adjust (yet)
                pass
            # Store new variable
            variables[key] = value

        # Close file
        f.close()

        self.stations = stations
        self.variables = variables
        self.data = data



"""This module contains methods for diagnosing the time series models for
its statistic assumptions.


Autocorrelation:
    - durbin_watson
    - ljung_box
    - runs_test
    - breusch_godfrey



"""

from numpy import abs, sqrt, cumsum
from pandas import DataFrame, concat
from scipy.stats import chi2, norm

from .core import acf as get_acf

__all__ = ["ljung_box", "runs_test", "durbin_watson", ]


def durbin_watson(series=None, acf=None, alpha=0.05, **kwargs):
    """Method to calculate the durbin watson statistic.

    Parameters
    ----------
    series: pandas.Series
        the autocorrelation function.
    tmin: str
    tmax: str
    kwargs:
        all keyword arguments are passed on to the acf function.

    Returns
    -------
    DW: float

    Notes
    -----
    The Durban Watson statistic can be used to make a statement on the
    correlation between the values. The formula to calculate the Durbin
    Watson statistic (DW) is:

    .. math::

        DW = 2 * (1 - acf(s))

    where acf is the autocorrelation of the series for lag s. By
    definition, the value of DW is between 0 and 4. A value of zero
    means complete negative correlation and 4 indicates complete
    positive autocorrelation. A value of zero means no autocorrelation.

    References
    ----------
    .. [DW} Durbin, J., & Watson, G. S. (1951). Testing for serial correlation
    in least squares regression. II. Biometrika, 38(1/2), 159-177.

    .. [F] Fahidy, T. Z. (2004). On the Application of Durbin-Watson
    Statistics to Time-Series-Based Regression Models. CHEMICAL ENGINEERING
    EDUCATION, 38(1), 22-25.

    TODO
    ----
    Compare calculated statistic to critical values, which are
    problematic to calculate and should come from a predefined table.

    """
    if acf is None:
        acf = get_acf(series, **kwargs)

    DW = 2 * (1 - acf)
    DW.name = "DWtest"

    # result = DataFrame(index=lags, data={"Qtest": Qtest, "P-value": pval})
    # result.index.name = "Lags (Days)"
    # name = "Accept Ha (alpha={})".format(alpha)
    # result[name] = pval < alpha

    return DW


def ljung_box(series=None, acf=None, alpha=0.05, **kwargs):
    """Ljung-box test for autocorrelation. Either a "series" or "acf" has to be
    provided.

    Parameters
    ----------
    series: pandas.Series, optional
        series to calculate the autocorrelation for that is used in the
        Ljung-Box test.
    acf: pandas.Series, optional
        The autocorrelation function used in the Ljung-Box test. Using a
        pre-calculated acf will be faster.
    alpha: float, optional
        Significance level to test against. Float values between 0 and 1.
        Default is alpha=0.05.
    kwargs:
        The keyword arguments provided to this method will be passed on the
        the ps.stats.acf method.

    Returns
    -------
    h: bool
        True if series has significant autocorrelation for any of the lags.
    Pandas.DataFrame
        DataFrame containing the Q test statistic and the p-value.

    Notes
    -----
    The Ljung-Box test can be used to test autocorrelation in the
    residuals series which are used during optimization of a model. The
    Ljung-Box Q-test statistic is calculated as :

    .. math::

        Q(k) = nobs * (n + 2) * \Sum(acf^2(k) / (n - k)

    where `k` are the lags to calculate the autocorrelation for,
    nobs is the number of observations and `acf(k)` is the autocorrelation for
    lag `k`. The Q-statististic can be compared to the value of a
    Chi-squared distribution to check if the Null hypothesis (no
    autocorrelation) is rejected or not. The hypothesis is rejected when:

    .. math::

        Q(k) > Chi^2(\alpha, h)

    Where \alpha is the significance level and `h` is the degree of
    freedom defined by `h = nobs - p` where `p` is the number of parameters
    in the model.

    References
    ----------
    .. [LB] Ljung, G. and Box, G. (1978). "On a Measure of Lack of Fit in Time
    Series Models", Biometrika, 65, 297-303.

    Examples
    --------

    v = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
              data=np.random.rand(1000))
    ps.stats.acf(v)

    """
    if acf is None:
        acf = get_acf(series, **kwargs)

    # Drop zero-lag from the acf and drop nan-values
    acf.drop(0, inplace=True, errors="ignore")
    acf.dropna(inplace=True)

    lags = acf.index.values
    nobs = acf.index.size

    Qtest = nobs * (nobs + 2) * cumsum(acf.values ** 2 / (nobs - lags))

    # TODO decrease lags by number of parameters?
    pval = chi2.sf(Qtest, lags)

    result = DataFrame(index=lags, data={"LBtest": Qtest, "P-value": pval})
    result.index.name = "Lags (Days)"
    name = "Accept Ha (alpha={})".format(alpha)
    result[name] = pval < alpha
    h = result[name].any()

    return h, result


def runs_test(series, alpha=0.05, cutoff="mean"):
    """Runs test to test for autocorrelation. Returns true is there is
    significant autocorrelation according to this test.

    Parameters
    ----------
    series: pandas.Series
        Series to perform the runs test on.
    alpha: float, optional
        Significance level to use in the test.
    cutoff: str or float, optional
        String set to "mean" or "median" or a float to use as the cutoff.

    Returns
    -------
    h: bool
        Boolean that tells if the alternative hypothesis is accepted (True)
        or rejected (False) at confidence level alpha.
    Z: float
        Test-statistic
    pval: float
        p-value for the test statistic, based on a normal .

    Notes
    -----
    Ho: The series is a result of a random process
    Ha: The series is not the result of a random process

    If |Z| >= Z$_1-\alpha / 2$ then the null hypothesis (Ho) is rejected and
    the alternative hypothesis (Ha) is accepted.

    """
    # Make dichotomous sequence
    r = series.values.copy()
    if cutoff == "mean":
        cutoff = r.mean()
    elif cutoff == "median":
        cutoff = r.median()
    elif isinstance(cutoff, float):
        pass
    else:
        raise NotImplementedError("Cutoff criterion {} is not "
                                  "implemented".format(cutoff))

    r[r > cutoff] = 1
    r[r < cutoff] = 0

    # Calculate number of positive and negative noise
    n_pos = r.sum()
    n_neg = r.size - n_pos

    # Calculate the number of runs
    runs = r[1:] - r[0:-1]
    n_runs = sum(abs(runs)) + 1

    # Calculate the expected number of runs and the standard deviation
    n_neg_pos = 2.0 * n_neg * n_pos

    n_runs_exp = n_neg_pos / (n_neg + n_pos) + 1

    n_runs_std = (n_neg_pos * (n_neg_pos - n_neg - n_pos)) / \
                 ((n_neg + n_pos) ** 2 * (n_neg + n_pos - 1))

    # Calculate Z-statistic and pvalue
    Z = (n_runs - n_runs_exp) / sqrt(n_runs_std)
    pval = 2 * norm.sf(abs(Z))
    h = pval < alpha
    return h, Z, pval


def acf_test(series=None, acf=None, alpha=0.05, **kwargs):
    if acf is None:
        acf = get_acf(series, **kwargs)

    h, LB = ljung_box(acf=acf, alpha=alpha)
    DW = durbin_watson(acf=acf, alpha=alpha)
    h, _, _ = runs_test(series, alpha=alpha)

    report = concat([DW, LB], axis=1)

    return h, report


def breusch_godfrey(series=None, acf=None, alpha=0.05, **kwargs):
    return NotImplementedError("Method not implemented yet.")
    # result = DataFrame()
    # h = True
    # return h, result


def lilliefors(series, alpha=0.05, **kwargs):
    """Lilliefors test to test for normality of the time series

    Parameters
    ----------
    series
    alpha
    kwargs

    Returns
    -------
    h: bool
        boolean telling whether or not to reject the alternative hypothesis
        that the data is not normally distributed.
    """
    return NotImplementedError("Method not implemented yet.")
